///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// Copyright Nurve Networks LLC 2008
// 
// Filename: REM_AVR_NTSC_DOOM_FLOOR_Z_02.c
//
// Original Author: Remi Veilleux
// 
// Last Modified: 02.12.09

// Modified : 
// New version with texture stored in Flash (instead of SRAM)
// With different image for ceiling and floor
// And 8 levels of precalculated mip-mapping / shading
//
// Overview: fullscreen 'Doom-style' floor and ceiling
//
// Controls: gamepad, use d-pad (up/down, left/right) to move around, and button to look up/down
// The effect is running at a constant 60 frame per second refresh rate (NTSC)
// 
// 
// Demo must be linked with following source files:
// 
// REM_AVR_NTSC_DOOM_FLOOR_8_V011.s     - ASM NTSC bitmap driver.
//
// Note 1: header files for all sources must be in current working directory, so compiler can find them
//
// 
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES ///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// define chip type here
#define __AVR_ATmega644__

// define CPU frequency in Mhz here if not defined in Makefile, this is used for conditional compilation, so needs
// to be first, before other XGS includes
#define F_CPU 28636360UL

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <avr/eeprom.h>
#include <avr/io.h>
#include <avr/sleep.h>

#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <compat/twi.h>

// include local XGS API header files
#include "XGS_AVR_GAMEPAD_DRV_V010.h"

// include header file for video driver
#include "REM_AVR_NTSC_DOOM_FLOOR_8_V011.h"

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// TYPES  /////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
typedef unsigned short USHORT;
typedef unsigned short WORD;
typedef unsigned char  BYTE;
typedef unsigned char UCHAR;
typedef unsigned long  QUAD;
typedef unsigned int   UINT;
typedef signed long int INT32;
typedef signed int INT16;
typedef signed char INT8;

extern volatile UCHAR overscan_color;
extern UCHAR hscroll;
extern UCHAR vscroll;
extern UINT framecount;
extern UCHAR rv_avr_floor0;
extern UCHAR rv_avr_ceiling0;

// NOTE: these externs need to be declared volatile because they are modified by the asm interrupt driver
extern volatile unsigned int  curr_line;
extern volatile unsigned char curr_raster_line;

typedef struct
{
	int startx;
	int add_x;
	int starty;
	int add_y;
	UCHAR texture_address;
} SFloorLines;

extern UCHAR rv_avr_rotozoom64x32_v010[];

extern SFloorLines floorlines[];
extern INT16 costable[];
extern INT8 sintable[];

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// GLOBALS/////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#define SET_BIT(value, bit)    value = ( (value) | (1 << (bit)) ) 
#define SET(bit)               (1 << (bit))
#define RESET_BIT(value, bit)  value = ((value) &  (~(1 << (bit)))) 
#define RESET(bit)             (~(1 << (bit))))

int Gamepad_Init()
{
// this function intializes the gamepad, more or less simply sets the I/O ports in the correct direction, state, and
// pull ups. Must be called before using the read function, unless the I/O ports have been setup correctly elsewhere
// also, care must be taken not to disturb other I/O settings, so writes must be masked
// note the SETPORTBITS macro only sets bits, it can't reset them to 0!

// set data direction of gamepad I/O pins to receive inputs
RESET_BIT(DDRD, GAMEPAD_DATA0); 
RESET_BIT(DDRD, GAMEPAD_DATA1); 

// now enable pull ups on the inputs, so if controllers aren't connected there are no problems
SET_BIT(PORTD, GAMEPAD_DATA0); 
SET_BIT(PORTD, GAMEPAD_DATA1); 

// set data direction of gamepad control pins as outputs
SET_BIT(DDRD, GAMEPAD_SHnLD); 
SET_BIT(DDRC, GAMEPAD_CLK); 

// return success, later add some more relevant error codes
return(1);

} // end Gamepad_Init

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int Gamepad_Read(int gamepadio_pin)
{
// reads the value of the game pads and returns them in a single 16-bit value
volatile int button_state = 0;
volatile int counter = 0;

// step 1: make sure clock is low and latch is low
PORTD = PORTD & ~(1 << GAMEPAD_SHnLD);
PORTC = PORTC & ~(1 << GAMEPAD_CLK);

// step 2: now pulse latch HIGH/LOW to latch the data, leave in LOW which is shift mode
PORTD = PORTD | (1 << GAMEPAD_SHnLD);
PORTD = PORTD & ~(1 << GAMEPAD_SHnLD);

// step 4: shift remaining buttons out
for (counter = 0; counter < 7; counter++)
  {
  // insert the next bit into the lsb of buttons
  button_state = button_state | ((PIND & (1 << gamepadio_pin)) >> gamepadio_pin);

  // pulse clock
  PORTC = PORTC | (1 << GAMEPAD_CLK);
  PORTC = PORTC & ~(1 << GAMEPAD_CLK);

  // shift result and make space for next bit
  button_state = button_state << 1;

  } // end for

// step 3: get first bit and insert into button_state
button_state = button_state | ((PIND & (1 << gamepadio_pin)) >> gamepadio_pin);

// return results
return ((~button_state) & 0xFF);

} // end Gamepad_Read


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MAIN ///////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


void GFX_Init(int video_interrupt_rate)
{
	// this function initializes the I/O ports for NTSC/PAL/VGA graphics as well as initilizes variables and turns on the 
	// driver interrupt which will immediately start feeding the data sources to the currently installed ASM driver, thus
	// make sure any pointers to data structures are initialized before calling this function
	// when done with graphics and you want to stop calling the graphics interrupt, call GFX_Shutdown(..)
	//
	// Inputs:
	//     video_interrupt_rate - the clock divider, so that the number of clocks per interrupt is equal to the HSYNC rate
	//                            since the interrupt handler runs every HSYNC in both NTSC/PAL and VGA modes

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// VIDEO INTERRUPT EXPERIMENT /////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// main graphics port that luma/chroma and VGA signals come from, port A
	#define GFX_MAIN_PORT       PORTA
	#define GFX_MAIN_DDR        DDRA

	// the "saturation" control is connected to port B
	#define GFX_SATURATION_PORT PORTB
	#define GFX_SATURATION_DDR  DDRB

	// set video ports to output
	GFX_MAIN_DDR  = 0xFF; // NTSC/PAL LUMA7:4 | CHROMA3:0, VGA7:0 [vsync, hsync, B1,B0, G1,G0, G1,G0] 
	              //                                        b7      b6    b5:4   b3:2   b1:0
	GFX_MAIN_PORT = 0x00;

	#define SETPORTBITS(b7, b6, b5, b4, b3, b2, b1, b0)  ( (b7 << 7) | (b6 << 6)| (b5 << 5)| (b4 << 4)| (b3 << 3)| (b2 << 2)| (b1 << 1)| (b0 << 0) )

	// set sat control bits PB2, PB3 as outputs
	GFX_SATURATION_DDR  =  SETPORTBITS(0,0,0,0,1,1,0,0);
	PORTB               =  SETPORTBITS(0,0,0,0,1,1,0,0);

	// set up timer                                     
	TCCR1A = (0 << COM1A1) | (0 << COM1A0) | (0 << COM1B1) | (0 << COM1B0) | (0 << WGM11) | (0 << WGM10);

	// set clock pre-scaler and counter mode
	TCCR1B = (0 << WGM13) | (1 << WGM12) | (0 << CS12) | (0 << CS11)  | (1 << CS10);    

	// 168. 644 use TIMSK1,TIFR1
	// 162 use TIMSK,TIFR

	// enable timer interrupts for timer 
	TIMSK1 = (1 << OCIE1A);                  //(1<<TOIE1);                 

	// clear and pendings timer x CTC interrupts (1 clears it)
	TIFR1  = (1 << OCF1A); // (1<<TOV1);                  

	// set output compare register value
	OCR1A  = video_interrupt_rate; // 1817+-1 for NTSC at 28.636 Mhz = 15.748Hkz hsync rate, 910 for VGA @ 28.636Mhz = 31.468Khz, hsync rate

	// enable interrupt
	sei();
} // end GFX_Init

inline void RotateVectors16(INT16* ioVect0, INT16* ioVect1, INT16 inRotAngle)
{
	INT16 oldx = *ioVect0;
	INT16 a = inRotAngle & 16383;
	if(a < 0) { a = 16384 + a; }
	INT16 cosa = (INT16) pgm_read_word(costable + a);
	INT16 sina = (INT16) pgm_read_word(costable + ((a - 4096) & 16383));
	*ioVect0 = ((INT32)*ioVect0 * cosa - (INT32)*ioVect1 * sina) >> 15;
	*ioVect1 = ((INT32)oldx * sina + (INT32)*ioVect1 * cosa) >> 15;
}

INT16 mPosX = 2000;
INT16 mPosY = 2000;
INT16 mPosZ = -120;
INT16 mBasePosZ = 190; 
INT16 mAngle = 2200;
INT16 mDesiredAngle = 2200;
INT16 mRoll = 70;
INT16 mCurrentSpeed0 = 0;
INT16 mCurrentSpeed1 = 0;
INT16 mSpeedX = 0;
INT16 rotate_speed = 80;

int main(void)
{
	UCHAR j;
	int gamepad;           // used to read gamepad
	INT16 moveangle;
	UCHAR moveSpeed;
	INT16 trypos0;
	INT16 trypos1;
	INT16 mCameraX0 = ((INT32)(NTSC_SCREEN_WIDTH) << 11) / (INT32)NTSC_SCREEN_WIDTH - 1024; //x-coordinate in camera space
	INT16 mCameraX1 = -1024;
	INT16 mPlaneYSize = 3435 >> 2;
	//INT16 mInvScreenW = (INT32)65536 / (INT32)NTSC_SCREEN_WIDTH;
	INT16 mDir0;
	INT16 mDir1;
	INT16 mPlane0;
	INT16 mPlane1;
	INT16 mHorizonCenter;
	// eye distance controls the 'zoom factor' of the texture
	INT16 mEyeDistance = (((INT32)NTSC_SCREEN_WIDTH) << 11) / 2435;
	INT16 ceil_int;
	INT16 floor_int;
	INT16 diff;
	INT16 rayDirX;
	INT16 rayDirY;
	INT16 end_rayDirX;
	INT16 end_rayDirY;
	INT16 distance_int;
	INT16 startx;
	INT16 starty;

	// initialize the video driver and GFX engine
	GFX_Init( VIDEO_INTERRUPT_RATE );

	// initialize gamepad
	Gamepad_Init();

	// set processor sleep mode
	set_sleep_mode( SLEEP_MODE_IDLE );
	sleep_enable();

	vscroll = 30;
	hscroll = 28;

	// main event loop
	while(1)
	{
	    // main loop that executes and is interrupted by video interrupt

		overscan_color = NTSC_BLACK;
	    // now goto sleep and wait for interrupt
	    sleep_cpu();

	    // only do video processing on a single line (at the top of the video scan)
	    if ( curr_line < END_ACTIVE_SCAN ) continue;
	    
		//overscan_color = NTSC_ORANGE;
		framecount++;
       	gamepad = Gamepad_Read(GAMEPAD0);


		if(gamepad & GAMEPAD_RIGHT)
		{
			mDesiredAngle += rotate_speed;
		}
		else if(gamepad & GAMEPAD_LEFT)
		{
			mDesiredAngle -= rotate_speed;
		}
		mAngle = ((INT32)mAngle*230 + (INT32)mDesiredAngle*26)>>8;
	
		if((gamepad & GAMEPAD_ACTION1) && mRoll < 150)
		{
			mRoll += 3;
			mBasePosZ += 2;
		}
		if((gamepad & GAMEPAD_ACTION2) && mRoll > -150)
		{
			mRoll -= 3;
			mBasePosZ -= 2;
		}

		trypos0 = -1024;
		trypos1 = 0;
		moveangle = -1;
		moveSpeed = 2;
		if(gamepad & GAMEPAD_UP) moveangle = 0;
		if(gamepad & GAMEPAD_DOWN) moveangle = 8192;
		mPosZ = ((mPosZ * 15) + mBasePosZ) >> 4;
		if(moveangle != -1)
		{
			RotateVectors16(&trypos0, &trypos1, mAngle + moveangle);
			trypos0 *= moveSpeed;
			trypos1 *= moveSpeed;
		
			mPosZ += ((INT16)pgm_read_word(costable + (((framecount<<9)&16383))) >> 12);
		}
		else
		{
			trypos0 = 0;
			trypos1 = 0;
		}

		mCurrentSpeed0 = ((INT32)mCurrentSpeed0*7 + trypos0*1) / 8;
		mCurrentSpeed1 = ((INT32)mCurrentSpeed1*7 + trypos1*1) / 8;
	    mPosX += mCurrentSpeed0>> 6;
	    mPosY += mCurrentSpeed1>> 6;

		mHorizonCenter = ((NTSC_SCREEN_HEIGHT >> 1) - ((mRoll * NTSC_SCREEN_HEIGHT) >> 8));
		ceil_int = ((325-mPosZ) * mEyeDistance);
		floor_int = (mPosZ * mEyeDistance);

		mDir0 = -1024;
		mDir1 = 0;
		mPlane0 = 0;
		mPlane1 = mPlaneYSize;
		RotateVectors16(&mDir0, &mDir1, mAngle);
		RotateVectors16(&mPlane0, &mPlane1, mAngle);
		
		rayDirX = mDir0 + (((INT32)mPlane0 * mCameraX0) >> 10);
		rayDirY = mDir1 + (((INT32)mPlane1 * mCameraX0) >> 10);
		end_rayDirX = mDir0 + (((INT32)mPlane0 * mCameraX1) >> 10);
		end_rayDirY = mDir1 + (((INT32)mPlane1 * mCameraX1) >> 10);

		//overscan_color = NTSC_GREEN;
		// low detail mode: render one scanline out of two, alternate between frames
		if(framecount&1)
		{
			j=1;
		}
		else
		{
			j=0;
		}
		for(; j<NTSC_SCREEN_HEIGHT; j+=2)
		{
		
			diff = j-mHorizonCenter;
			if(diff <= 0)
			{
				diff = 1-diff;
				distance_int = (ceil_int / diff);
				floorlines[j].texture_address = (UINT)(&rv_avr_ceiling0)>>8;
			}
			else
			{
				distance_int = (floor_int / diff);
				floorlines[j].texture_address = (UINT)(&rv_avr_floor0)>>8;
			}

			// Each texture is 32x32 = 1024 bytes, so hi8() is +4 from previous shading
			if(distance_int > 3400) floorlines[j].texture_address += 28;
			else if(distance_int > 2200) floorlines[j].texture_address += 24;
			else if(distance_int > 1000) floorlines[j].texture_address += 20;
			else if(distance_int > 680) floorlines[j].texture_address += 16;
			else if(distance_int > 460) floorlines[j].texture_address += 12;
			else if(distance_int > 350) floorlines[j].texture_address += 8;
			else if(distance_int > 210) floorlines[j].texture_address += 4;

			startx = (((INT32)rayDirX * distance_int) >> 8);
			starty = (((INT32)rayDirY * distance_int) >> 8);
		
			//floorlines[j].add_x = (((((INT32)end_rayDirX * distance_int) >> 8)-startx) * mInvScreenW) >> 13;
			//floorlines[j].add_y = (((((INT32)end_rayDirY * distance_int) >> 8)-starty) * mInvScreenW) >> 13;
			// Cheap optimisation: mInvScreenW=744 so it's somewhat close to 1024
			// thus <<10>>13 can be reduced to >>3
			// However floor perspective will get displayed a little bit off-centered
			floorlines[j].add_x = (((((INT32)end_rayDirX * distance_int) >> 8)-startx) ) >> 3;
			floorlines[j].add_y = (((((INT32)end_rayDirY * distance_int) >> 8)-starty) ) >> 3;
			floorlines[j].startx = (mPosX + startx) << 3;
			floorlines[j].starty = (mPosY + starty) << 3;
		}
      
	} // end while
} // end main

