///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// Copyright Nurve Networks LLC 2008
// 
// Filename: REM_AVR_ULTRA_DEMO_NTSC_01.c
//
// Original Author: Remi Veilleux
// 
// Last Modified: 01.09.09
//
// Overview: Unfinished technology demo to showcase Rem's Ultra graphic engine
//
// Controls: gamepad, use d-pad to scroll background around
// 
// Demo must be linked with following source files:
// 
// XGS_AVR_SYSTEM_V010.c              - System variables.
// REM_AVR_NTSC_128_96_ULTRA_V010.s   - ASM NTSC bitmap driver.
// XGS_AVR_GAMEPAD_DRV_V010.c         - Gamepad driver.
//
// Note 1: header files for all sources must be in current working directory, so compiler can find them
//
// 
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES ///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// define chip type here
#define __AVR_ATmega644__

// define CPU frequency in Mhz here if not defined in Makefile, this is used for conditional compilation, so needs
// to be first, before other XGS includes, also note there is a copy of this constant in XGS_AVR_SYSTEM_V010, so if you 
// change it here, also change it in the header to make certain everthing is sync'ed
#define F_CPU 28636360UL

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <avr/eeprom.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>

#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <compat/twi.h>

// include local XGS API header files
#include "XGS_AVR_SYSTEM_V010.h"
#include "XGS_AVR_GAMEPAD_DRV_V010.h"

// include header file for video driver
#include "REM_AVR_NTSC_128_96_ULTRA_V010.h"

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// TYPES  /////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
extern UCHAR overscan_color;
extern UCHAR hscroll;
extern UCHAR vscroll;
extern UINT framecount;
extern UCHAR* tilemap[];
extern UCHAR rem_avr_ultra_tiles;
extern UCHAR sintable[];
typedef struct
{
	UCHAR rowindex;
	UCHAR xoffset;
} SScanList;
extern SScanList scanlist[];
extern prog_uchar jmptable;

typedef struct
{
	UCHAR address;
	UCHAR x;
} SSprite;
extern SSprite spritelist[];


// NOTE: these externs need to be declared volatile because they are modified by the asm interrupt driver
extern volatile unsigned int  curr_line;
extern volatile unsigned char curr_raster_line;


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// GLOBALS/////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MAIN ///////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


void GFX_Init(int video_interrupt_rate)
{
	// this function initializes the I/O ports for NTSC/PAL/VGA graphics as well as initilizes variables and turns on the 
	// driver interrupt which will immediately start feeding the data sources to the currently installed ASM driver, thus
	// make sure any pointers to data structures are initialized before calling this function
	// when done with graphics and you want to stop calling the graphics interrupt, call GFX_Shutdown(..)
	//
	// Inputs:
	//     video_interrupt_rate - the clock divider, so that the number of clocks per interrupt is equal to the HSYNC rate
	//                            since the interrupt handler runs every HSYNC in both NTSC/PAL and VGA modes

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// VIDEO INTERRUPT EXPERIMENT /////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// main graphics port that luma/chroma and VGA signals come from, port A
	#define GFX_MAIN_PORT       PORTA
	#define GFX_MAIN_DDR        DDRA

	// the "saturation" control is connected to port B
	#define GFX_SATURATION_PORT PORTB
	#define GFX_SATURATION_DDR  DDRB

	// set video ports to output
	GFX_MAIN_DDR  = 0xFF; // NTSC/PAL LUMA7:4 | CHROMA3:0, VGA7:0 [vsync, hsync, B1,B0, G1,G0, G1,G0] 
	              //                                        b7      b6    b5:4   b3:2   b1:0
	GFX_MAIN_PORT = 0x00;

	// set sat control bits PB2, PB3 as outputs
	GFX_SATURATION_DDR  =  SETPORTBITS(0,0,0,0,1,1,0,0);
	PORTB               =  SETPORTBITS(0,0,0,0,1,1,0,0);

	// set up timer                                     
	TCCR1A = (0 << COM1A1) | (0 << COM1A0) | (0 << COM1B1) | (0 << COM1B0) | (0 << WGM11) | (0 << WGM10);

	// set clock pre-scaler and counter mode
	TCCR1B = (0 << WGM13) | (1 << WGM12) | (0 << CS12) | (0 << CS11)  | (1 << CS10);    

	// 168. 644 use TIMSK1,TIFR1
	// 162 use TIMSK,TIFR

	// enable timer interrupts for timer 
	TIMSK1 = (1 << OCIE1A);                  //(1<<TOIE1);                 

	// clear and pendings timer x CTC interrupts (1 clears it)
	TIFR1  = (1 << OCF1A); // (1<<TOV1);                  

	// set output compare register value
	OCR1A  = video_interrupt_rate; // 1817+-1 for NTSC at 28.636 Mhz = 15.748Hkz hsync rate, 910 for VGA @ 28.636Mhz = 31.468Khz, hsync rate

	// enable interrupt
	sei();
} // end GFX_Init

int main(void)
{
	int gamepad;           // used to read gamepad
	int lastgamepad;
	int i, j, k;
	UCHAR c1, c2;

	#define MAXBALLS 36
	struct
	{
		UINT x, y;
		int dx, dy;
	} balls[MAXBALLS]; // Struct takes 8 bytes * MAXBALLS = 288 of stack SRAM

	// Prepare address for tiles
	for(j=0; j<13; j++)
	{
		for(i=0; i<TILEMAP_WIDTH; i++)
		{
			tilemap[(j*TILEMAP_WIDTH) + i] = &rem_avr_ultra_tiles + (i*8) + (j*256*8);
		}
	}
	// Build a 'overlay screen' by using tile 24/25/26 and 27/28/29
	for(i=0; i<TILEMAP_WIDTH; i++)
	{
		tilemap[(13*TILEMAP_WIDTH) + i] = &rem_avr_ultra_tiles + (24*8);
		tilemap[(14*TILEMAP_WIDTH) + i] = &rem_avr_ultra_tiles + (25*8);
	}
	tilemap[(14*TILEMAP_WIDTH) + 3] = &rem_avr_ultra_tiles + (26*8);
	tilemap[(14*TILEMAP_WIDTH) + 5] = &rem_avr_ultra_tiles + (26*8);


	for(i=0; i<MAXBALLS; i++)
	{
		balls[i].x = (rand()&127)<<8;
		balls[i].y = (rand()%78)<<8;
		balls[i].dx = (rand()&511) - 256;
		balls[i].dy = 0;
	}

	// Initialize scanlist: this tells which tilemap scanline gets rendered by each scanline.
	for(i=0; i<192; i++)
	{
		scanlist[i].rowindex = i>>1;
		scanlist[i].xoffset = 0;
	}
	// Special case: this last line should always be 'disabled' and point to black
	// We'll make it point to line 7 of tilerow 13
	scanlist[191].rowindex = 13*8+7;

	vscroll = 0;
	hscroll = 1;
/*
	for(j=0; j<96; j++)
	{
		spritelist[j*8].x = 8;
		//&jmptable doesn't work: I can't get the correct address from C. 
		// address should be : 0x0396
		spritelist[j*8].address = (j&7)*2;
		if(j>=48) spritelist[j*8].address += 16;

		spritelist[j*8+1].address = ((j&7)+8)*2;
		spritelist[j*8+1].x = 50;

		spritelist[j*8+2].address = ((j&7)+8)*2;
		spritelist[j*8+2].x = 70;

		spritelist[j*8+3].address = ((j&7)+8)*2;
		spritelist[j*8+3].x = 80;

		spritelist[j*8+4].address = ((j&7)+8)*2;
		spritelist[j*8+4].x = 90;

		spritelist[j*8+5].address = ((j&7)+8)*2;
		spritelist[j*8+5].x = 100;

		spritelist[j*8+6].address = ((j&7)+8)*2;
		spritelist[j*8+6].x = 110;

		spritelist[j*8+7].address = ((j&7))*2;
		spritelist[j*8+7].x = 120;
	}
*/
	// initialize the video driver and GFX engine
	GFX_Init( VIDEO_INTERRUPT_RATE );

	// initialize gamepad
	Gamepad_Init();

	// set processor sleep mode
	set_sleep_mode( SLEEP_MODE_IDLE );
	sleep_enable();
	//overscan_color = NTSC_ORANGE;

	// main event loop
	while(1)
	{
	    // main loop that executes and is interrupted by video interrupt

	    // now goto sleep and wait for interrupt
	    sleep_cpu();

	    // only do video processing on a single line (at the end of the video active scan)
	    if ( curr_line != END_ACTIVE_SCAN ) continue;

		// From here on, we have about 100,000 clock cycle before next frame begins
		// overscan_color (for debugging) will show the used portion
		// starting with bottom overscan, then vblank, then top overscan.

		// Enabled this line for debugging: to visualize how much time this processing code takes
		//overscan_color = NTSC_RED;

        gamepad = Gamepad_Read(GAMEPAD0);
		
		// Enabled this to test gamepad input more precisely
		/*
		do
		{
			if(gamepad)
			{
				if(gamepad == lastgamepad)
				{
					gamepad = 0;
					break;
				}
			}
			lastgamepad = gamepad;
		} while(0);
		*/

		if ((gamepad & GAMEPAD_UP) && vscroll > 0) vscroll--;
		if ((gamepad & GAMEPAD_DOWN) && vscroll < (160-96)*2) vscroll++;
		if ((gamepad & GAMEPAD_LEFT) && hscroll > 0) hscroll--;
		if ((gamepad & GAMEPAD_RIGHT) && hscroll < 64) hscroll++;
/*
		for(j=0; j<TILEMAP_HEIGHT; j++)
		{
			for(i=0; i<TILEMAP_WIDTH; i++)
			{
				tilemap[(j << 4) + i] = &tiles_256x8 + (((i+j)%3)<<3);
			}
		}*/

		/*
		for(j=0; j<96; j++)
		{
			//t2 = pgm_read_byte_near(sintable+((j+t)&255))>>2;
			spritelist[j*8].x = hscroll + ((pgm_read_byte_near(sintable+(((framecount<<1)+(j<<3))&255)))>>4);
			//spritelist[j*8].address = (((pgm_read_byte_near(sintable+(t2+(j<<3)&255))>>5)+(t>>1))&7)*2;
		}
		*/

	// Animate balls
	
	{
		for(j=0; j<MAXBALLS; j++)
		{
			balls[j].x+=balls[j].dx;
			if(balls[j].x >= (136<<8))
			{
				balls[j].x = (136<<8);
				balls[j].dx = -(rand()&511)-50;
			}
			else if(balls[j].x < 256)
			{
				balls[j].x = 256;
				balls[j].dx = (rand()&511)+50;
			}

			balls[j].dy+=13;
			balls[j].y+=balls[j].dy;

			if(balls[j].y > (87<<8))
			{
				balls[j].y = (87<<8);
				balls[j].dy = -balls[j].dy;
			}
		}
	}
	
	// Clear display list
	{
		for(j=0; j<96*8; j++)
		{
			spritelist[j].x = 0;
		}
	}

	{
		// For each ball, find an empty sprite in display list (if any!)
		for(j=0; j<MAXBALLS; j++)
		{
			// For each scanline of a ball
			for(k=0; k<8; k++)
			{
				c1 = (balls[j].y>>8) + k;

				for(i=0; i<8; i++)
				{
					if(spritelist[c1*8 + i].x != 0) continue;

					spritelist[c1*8 + i].x = balls[j].x>>8;
					spritelist[c1*8 + i].address = ((k&7)) + ((j&31)<<3);
					break;
				}
			}
		}
	}

	{
		c1 = vscroll&15;
		for(i=0; i<191; i++)
		{
			scanlist[i].rowindex = (((i+c1)>>1));
			//hscroll = (pgm_read_byte_near(sintable + ((/*(framecount<<2)+*/(i<<1))&255))) >> 5;
			scanlist[i].xoffset = hscroll&7;
		}
		c1 = vscroll>>4;
		c2 = hscroll>>3;
		for(j=0; j<13; j++)
		{
			for(i=0; i<TILEMAP_WIDTH; i++)
			{
				tilemap[(j*TILEMAP_WIDTH) + i] = &rem_avr_ultra_tiles + ((i+c2)*8) + ((j+c1)*256*8);
			}
		}

		// Make 'infinite scroll' in status screen
		c1 = framecount&7;
		// set a 'floating' status screen using tilerow 14 and 15
		for(i=60; i<=75; i++)
		{
			scanlist[i].xoffset = c1;
		}
		// Every 8 pixels, scroll tilemap and add a tile
		
		if(c1 == 0)
		{
			for(i=0; i<TILEMAP_WIDTH-1; i++)
			{
				tilemap[(14*TILEMAP_WIDTH) + i] = tilemap[(14*TILEMAP_WIDTH) + i+1];
			}
			tilemap[(14*TILEMAP_WIDTH) + TILEMAP_WIDTH-1] = &rem_avr_ultra_tiles + (rand()&31)*8;
		}
		

		scanlist[60].rowindex = 13*8 + 0;
		scanlist[61].rowindex = 13*8 + 1;
		scanlist[62].rowindex = 13*8 + 2;
		scanlist[63].rowindex = 13*8 + 3;
		scanlist[64].rowindex = 14*8 + 0;
		scanlist[65].rowindex = 14*8 + 1;
		scanlist[66].rowindex = 14*8 + 2;
		scanlist[67].rowindex = 14*8 + 3;
		scanlist[68].rowindex = 14*8 + 4;
		scanlist[69].rowindex = 14*8 + 5;
		scanlist[70].rowindex = 14*8 + 6;
		scanlist[71].rowindex = 14*8 + 7;
		scanlist[72].rowindex = 13*8 + 4;
		scanlist[73].rowindex = 13*8 + 5;
		scanlist[74].rowindex = 13*8 + 6;
		scanlist[75].rowindex = 13*8 + 7;

		// debug: display a row of debugging tiles (multicolor bars)
		/*
		for(i=0; i<TILEMAP_WIDTH; i++)
		{
			tilemap[(3*TILEMAP_WIDTH) + i] = &rem_avr_ultra_tiles + (27*8);
		}
		*/

		// Hide all sprites from these scanlines: this will make the 'floating screen'
		// appear 'above' sprites!
		for(j=60>>1; j<76>>1; j++)
		{
			for(i=0; i<8; i++)
			{
				spritelist[j*8 + i].x = 0;
			}
		}
	}

		framecount++;
		//Delay_Clocks(17000); // Warning: this only works if optimisation is disabled (-O0).
		// With optimisation -O2, the compiler skips it entirely.
		overscan_color = NTSC_BLACK;
  
	} // end while
} // end main

