///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// Copyright Nurve Networks LLC 2008
// 
// Filename: XGS_AVR_GAMEPAD_V010.c
//
// Original Author: Andre' LaMothe
// 
// Last Modified: 8.28.08
//
// Overview: The game pads are Nintendo Entertainment Systems (NES) based technology. They are slightly different in
// physical representation as well as these controllers have female DB9 rather than the proprietary NES connectors.
// Each controller is the same, so reading each is similar. The controllers use a simple serial encoding scheme where
// the button state is encoded as a stream of 8-bits. To read the game controller state, the caller simply latches
// the data state, then shifts out the button states bit by bit via a clock signal applied to the clock pin.
// The interface is shown below:
//
// I/O pin map for game controllers
// 
// GAMEPAD_DATAOUT_0 | PD4 | Pin 13 | Output
// GAMEPAD_DATAOUT_1 | PD5 | Pin 14 | Output
// GAMEPAD_SH/LDn    | PD6 | Pin 15 | Input
// GAMEPAD_CLK       | PC6 | Pin 26 | Input
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES ///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// this processor include needs to happen in the master file
// #define __AVR_ATmega644__

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <avr/eeprom.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <compat/twi.h>
#include <avr/pgmspace.h>

// include local XGS API header files
#include "XGS_AVR_SYSTEM_V010.h"
#include "XGS_AVR_GAMEPAD_DRV_V010.h"

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MACROS /////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DEFINES/////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// TYPES/CLASSES //////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// GLOBALS ////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// EXTERNALS //////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// PROTOTYPES /////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FUNCTIONS  /////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int Gamepad_Init()
{
// this function intializes the gamepad, more or less simply sets the I/O ports in the correct direction, state, and
// pull ups. Must be called before using the read function, unless the I/O ports have been setup correctly elsewhere
// also, care must be taken not to disturb other I/O settings, so writes must be masked
// note the SETPORTBITS macro only sets bits, it can't reset them to 0!

// set data direction of gamepad I/O pins to receive inputs
RESET_BIT(DDRD, GAMEPAD_DATA0); 
RESET_BIT(DDRD, GAMEPAD_DATA1); 

// now enable pull ups on the inputs, so if controllers aren't connected there are no problems
SET_BIT(PORTD, GAMEPAD_DATA0); 
SET_BIT(PORTD, GAMEPAD_DATA1); 

// set data direction of gamepad control pins as outputs
SET_BIT(DDRD, GAMEPAD_SHnLD); 
SET_BIT(DDRC, GAMEPAD_CLK); 

// return success, later add some more relevant error codes
return(1);

} // end Gamepad_Init

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int Gamepad_Read(int gamepadio_pin)
{
// reads the value of the game pads and returns them in a single 16-bit value
volatile int button_state = 0;
volatile int counter = 0;

// step 1: make sure clock is low and latch is low
PORTD = PORTD & ~(1 << GAMEPAD_SHnLD);
PORTC = PORTC & ~(1 << GAMEPAD_CLK);

// step 2: now pulse latch HIGH/LOW to latch the data, leave in LOW which is shift mode
PORTD = PORTD | (1 << GAMEPAD_SHnLD);
PORTD = PORTD & ~(1 << GAMEPAD_SHnLD);

// step 4: shift remaining buttons out
for (counter = 0; counter < 7; counter++)
  {
  // insert the next bit into the lsb of buttons
  button_state = button_state | ((PIND & (1 << gamepadio_pin)) >> gamepadio_pin);

  // pulse clock
  PORTC = PORTC | (1 << GAMEPAD_CLK);
  PORTC = PORTC & ~(1 << GAMEPAD_CLK);

  // shift result and make space for next bit
  button_state = button_state << 1;

  } // end for

// step 3: get first bit and insert into button_state
button_state = button_state | ((PIND & (1 << gamepadio_pin)) >> gamepadio_pin);

// return results
return ((~button_state) & 0xFF);

} // end Gamepad_Read


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// TWI/I2C INTERFACE //////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

