#define WIN32_LEAN_AND_MEAN		// Exclude rarely-used stuff from Windows headers
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include <string.h>

// Disable microsoft useless security warnings
#pragma warning(disable : 4996)

#define uint32 unsigned int
#define R_SHIFT 16
#define G_SHIFT 8
#define B_SHIFT 0
#define COLOR(r,g,b) (((r)<<R_SHIFT)+((g)<<G_SHIFT)+((b)<<B_SHIFT))
#define GETCOLOR(c,r,g,b) (r=(c>>R_SHIFT)&255,g=(c>>G_SHIFT)&255,b=(c>>B_SHIFT)&255)
template<typename T> T minmax(T a, T min, T max) { return (a < min ? min : (a > max ? max : a)); }

enum EConvertMode
{
	MODE_NORMAL,
	MODE_SPRITE
};

void convert(const char* in_FileName, EConvertMode in_Mode, int dither_level);
void readBMP(const char* in_FileName, int& out_width, int& out_height, uint32*& out_mem);
int getXGSColor(uint32 col);
void sintable();

// Globals:
int nbrefcolor; // number of reference color in reference BMP
uint32* referencecolor; // pointer to reference color

char st[4096];

int main(int argc, char* argv[])
{
	printf("XGS_utils: by Remi Veilleux\n");

	if(argc == 2)
	{
		if(stricmp(argv[1], "-sin") == 0)
		{
			sintable();
			return 0;
		}
	}
	if(argc >= 3)
	{
		if(stricmp(argv[1], "-s") == 0)
		{
			// For sprite, it is best to turn off dithering because we use pure black as transparency color
			convert(argv[2], MODE_SPRITE, 0);
		}
		if(stricmp(argv[1], "-i") == 0)
		{
			int dither_level = 800;
			if(argc == 4) dither_level = atoi(argv[3]);
			convert(argv[argc-1], MODE_NORMAL, dither_level);
		}
		return 0;
	}
	 
	MessageBox(NULL, "\
XGS_Utils is a command-line utility\n\
(but it can be used from Windows with a shortcut command argument)\n\
usage: xgs_utils [command] [image]\n\
  Various conversion and utils using xgs_color_reference_avr.bmp\n\
  Input: Windows 24-bit BMP (uncompressed)\n\
  Output: a file named 'something.h'\n\
  Available commands: -s: Compiled Sprite mode: output assembler code to render a compiled sprite\n\
                    -sin: Export a simple sintable (0..255 degree), 8-bit, unsigned\n\
                      -i: Image mode: output assembler code for this image\n\
                          Optional: -i [dither level] image_name\n\
                          Set the dithering level, from 0 (none) to 1000 (full). Default: 800"
	, "XGS_Utils", MB_ICONINFORMATION );
	
	return 0;
}

void sintable()
{
	// Generate sin table for XGS AVR:
	FILE* output = fopen("output.txt", "wt");
	if(!output)
	{
		printf("ERROR: Failed to open output file 'output.txt'\n");
		exit(1);
	}
	fprintf(output, "sintable:\n");
	char temp[2048];
	sprintf(temp, ".byte ");
	for(int i=0; i<256; i++)
	{
		int k = (int) (sinf(float(i/256.f)* 2.f * (float) M_PI) * 127) + 128;
		if((i&7) == 0)
		{
			printf("i=%i, k=%i, k>>5=%i\n", i, k, k>>5);
		}
		sprintf(st, "%i", k);
		strcat(temp, st);
		if(i < 255) strcat(temp, ",");
	}
	fprintf(output, "%s\n", temp);
	fclose(output);
}

void convert(const char* in_FileName, EConvertMode in_Mode, int dither_level)
{
	int width;
	int height;
	uint32* srcpic = NULL;
	readBMP(in_FileName, width, height, srcpic);
	
	int ref_width;
	int ref_height;
	readBMP("xgs_color_reference_avr.bmp", ref_width, ref_height, referencecolor);
	
	if(ref_height != 1)
	{
		printf("ERROR: Reference image should be a single pixel height\n");
		exit(1);
	}
	nbrefcolor = ref_width;
	
	int size = width * height;
	int* tempr = new int[size];
	int* tempg = new int[size];
	int* tempb = new int[size];
	
	// First pass: Split channels to separate int buffers
	int* bpoint = tempb;
	int* gpoint = tempg;
	int* rpoint = tempr;
	for(int j=0; j<height; j++)
	{
		uint32* src = srcpic + (j * width);
		for(int i=0; i<width; i++)
		{
			uint32 s = *src;
			int r,g,b;
			GETCOLOR(s, r, g, b);
			*bpoint++ = b;
			*gpoint++ = g;
			*rpoint++ = r;
			src++;
		}
	}
	
	bpoint = tempb;
	gpoint = tempg;
	rpoint = tempr;
	
	// This multiplication factor determines the amount of dithering:
	// 0 = no dithering at all
	// 1000 = full (100% dithering)
	int mult = dither_level;

	for(int j=0; j<height; j++)
	{
		uint32* dest = srcpic + (j * width);
		for(int i=0; i<width; i++)
		{
			// fetch source color (desired color)
			int r = *rpoint;
			int g = *gpoint;
			int b = *bpoint;
			
			// find closest color in reference table
			int bestdist = 999999;
			int bestk = 0;
			int r2, g2, b2;
			for(int k=0; k<nbrefcolor; k++)
			{
				GETCOLOR(referencecolor[k], r2, g2, b2);
				int dist = int(sqrtf(float((r2 - r)*(r2-r)*30 + (g2-g)*(g2-g)*59 + (b2-b)*(b2-b)*11)));

				if(dist < bestdist)
				{
					bestk = k;
					bestdist = dist;
				}
			}
			
			// Set this pixel to the selected color
			*dest = referencecolor[bestk];
			
			// Now diffuse error to surrounding pixels
			GETCOLOR(referencecolor[bestk], r2, g2, b2);
			for(int k=0; k<3; k++)
			{
				int error_diffusion;
				int* buffer;
				switch(k)
				{
					case 0: buffer = rpoint; error_diffusion = (r2 - r); break;
					case 1: buffer = gpoint; error_diffusion = (g2 - g); break;
					case 2: buffer = bpoint; error_diffusion = (b2 - b); break;
				}
				error_diffusion = error_diffusion * mult / 1000;
				
				error_diffusion = minmax(error_diffusion, -200, 200);
				if(error_diffusion != 0)
				{
					printf("Error diffusion=%i at pixel %i,%i\n", error_diffusion,i,j);
				}
				if(i<width-1) *(buffer+1) -= (error_diffusion*7)>>4;
				if(j<height-1)
				{
					if(i>0) *(buffer-1+width) -= (error_diffusion*3)>>4;
					*(buffer+width) -= (error_diffusion*5)>>4;
					if(i<width-1) *(buffer+1+width) -= (error_diffusion*1)>>4;
				}
			}
			
			dest++;
			rpoint++;
			gpoint++;
			bpoint++;
		}
	}
	
	char input_name[256];
	strcpy(input_name, in_FileName);
	char* dot = strchr(input_name, '.');
	if(!dot)
	{
		printf("ERROR: oops: input filename should end with .bmp '%s'\n", in_FileName);
		exit(1);
	}
	*dot = 0;
	
	char output_filename[256];
	sprintf(output_filename, "%s.h", input_name);
	
	FILE* output = fopen(output_filename, "wt");
	if(!output)
	{
		printf("ERROR: Can't create output file '%s'\n", output_filename);
		exit(1);
	}
	
	char temp[2048];
	if(in_Mode == MODE_NORMAL)
	{
		fprintf(output, "//XGS_Utils: converted from source image '%s'\n", in_FileName);
		fprintf(output, "//Image size is %ix%i\n", width, height);
		fprintf(output, ".global %s\n", input_name);
		fprintf(output, "%s:\n", input_name);
		
		for(int j=0; j<height; j++)
		{
			uint32* dest = srcpic + (j * width);
			sprintf(temp, ".byte ");
			for(int i=0; i<width; i++)
			{
				int value = getXGSColor(*dest++);
				
				sprintf(st, "0x%X", value);
				strcat(temp, st);
				if(i<width-1)
				{
					strcat(temp, ",");
				}
			}
			strcat(temp, "\n");
			fprintf(output, temp);
		}
		fprintf(output, "\n");
	}
	
	if(in_Mode == MODE_SPRITE)
	{
		int nbsprites = width / 8;
		fprintf(output, "//XGS_Utils: converted from source image '%s'\n", in_FileName);
		fprintf(output, "//to compiled sprite 8x%i, total of %i sprite(s)\n", height, nbsprites);
		
		int total_inst = 0;
		
		for(int k=0; k<nbsprites; k++)
		{
			int start_x = k*8;
			
			int count_inst = 0;
			for(int j=0; j<height; j++)
			{
				uint32* dest = srcpic + (j * width) + start_x;
				fprintf(output, "%s%i_row%i:\n", input_name, k, j);
				int count_clock = 0;
				int r16 = 0;
				for(int i=0; i<8; i++)
				{
					int value = getXGSColor(*dest++);
					if(value == 0x4F)
					{
					}
					else
					{
						if(r16 != value)
						{
							fprintf(output, "\tldi \tr16, 0x%0X\t// 1\n", value);
							count_clock += 1;
							count_inst += 1;
							r16 = value;
						}
						fprintf(output, "\tstd \tY+%i, r16\t// 2\n", i);
						count_clock += 2;
						count_inst += 1;
					}
				}
				if(count_clock > 24)
				{
					printf("ERROR: something wrong happened: sprite can't take more than 24 cycles!\n");
					exit(1);
				}
				// 'DELAY' takes a variable number of instruction so alas we can't have an exact count here (from 0 to 5)
				// Let's use '3' as average
				count_inst += 3;
				fprintf(output, "\tDELAYX\t%i\t// %i\n", 24-count_clock, 24-count_clock);
				count_inst += 1;
				fprintf(output, "\tret\n");
			}
			fprintf(output, "// Total bytes of FLASH for this sprite: %i\n", count_inst*2);
			total_inst += count_inst;
		}
		fprintf(output, "// Grand total bytes of FLASH for all sprites: %i\n", total_inst*2);
		
		fprintf(output, ".global jmptable\n");
		fprintf(output, "jmptable:\n");
		for(int k=0; k<nbsprites; k++)
		{
			for(int j=0; j<height; j++)
			{
				fprintf(output, "\tjmp \t%s%i_row%i\n", input_name, k, j);
			}
		}
	}
	
	fclose(output);
	
	printf("Done! Output file name is '%s'\n", output_filename);
}

int getXGSColor(uint32 col)
{
	int k;
	for(k=0; k<nbrefcolor; k++)
	{
		if(referencecolor[k] == col) break;
	}
	//_ASSERT(k<nbrefcolor);
	
	// Color section
	int chroma = (k/12);
	int luma = 0x40 + (k%12) * 0x10;
	return chroma+luma;
}

void readBMP(const char* in_FileName, int& out_width, int& out_height, uint32*& out_mem)
{
	FILE *input = fopen(in_FileName, "rb");
	if(!input)
	{
		printf("ERROR: Can't open file '%s'\n", in_FileName);
		exit(1);
	}
	char temp[256];
	short int bpp;
	short int planes;
	
	fread(temp, 1, 18, input); // skip first 18 bytes of header
	fread(&out_width, 1, 4, input);
	fread(&out_height, 1, 4, input);
	fread(&planes, 1, 2, input); // planes must be 1
	fread(&bpp, 1, 2, input); // bpp must be 24
	fread(temp, 1, 24, input); // skip junk
	
	printf("Reading image '%s', size=%ix%i, bpp=%i\n", in_FileName, out_width, out_height, bpp);
	
	if(planes != 1 || bpp != 24)
	{
		printf("ERROR: BMP must be a Windows 24-bit standard image\n");
		exit(1);
	}
	
	int size = out_width*out_height;
	out_mem = new uint32[size];
	memset(out_mem, 0, size*4);
	
	// Read and flip BMP because they are stored bottom-up
	for(int j=0; j<out_height; j++)
	{
		uint32* mem_row = out_mem + (out_height-1-j)*out_width;
		for(int i=0; i<out_width; i++)
		{
			fread(mem_row, 1, 3, input);
			mem_row++;
		}
	}
	
	fclose(input);
}
