/* db_expr.c - expression parser
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdlib.h>
#include "db_compiler.h"

/* local function prototypes */
static ParseTreeNode *ParseExpr2(ParseContext *c);
static ParseTreeNode *ParseExpr3(ParseContext *c);
static ParseTreeNode *ParseExpr4(ParseContext *c);
static ParseTreeNode *ParseExpr5(ParseContext *c);
static ParseTreeNode *ParseExpr6(ParseContext *c);
static ParseTreeNode *ParseExpr7(ParseContext *c);
static ParseTreeNode *ParseExpr8(ParseContext *c);
static ParseTreeNode *ParseExpr9(ParseContext *c);
static ParseTreeNode *ParseExpr10(ParseContext *c);
static ParseTreeNode *ParseExpr11(ParseContext *c);
static ParseTreeNode *ParsePrimary(ParseContext *c);
static ParseTreeNode *ParseSimplePrimary(ParseContext *c);
static ParseTreeNode *ParseArrayReference(ParseContext *c, ParseTreeNode *arrayNode);
static ParseTreeNode *ParseCall(ParseContext *c, ParseTreeNode *functionNode);
static ParseTreeNode *MakeUnaryOpNode(int op, ParseTreeNode *expr, Type *resultType);
static ParseTreeNode *MakeBinaryOpNode(int op, ParseTreeNode *left, ParseTreeNode *right, Type *resultType);
static ParseTreeNode *NewParseTreeNode(int type);

/* ParseRValue - parse and generate code for an r-value */
void ParseRValue(ParseContext *c)
{
    ParseTreeNode *expr;
    expr = ParseExpr(c);
    code_rvalue(c, expr);
    FreeParseTree(expr);
}

/* ParseExpr - handle the OR operator */
ParseTreeNode *ParseExpr(ParseContext *c)
{
    ParseTreeNode *node;
    Token tkn;
    node = ParseExpr2(c);
    if ((tkn = GetToken(c)) == T_OR) {
        ParseTreeNode *node2 = NewParseTreeNode(NodeTypeDisjunction);
        ExprListEntry *entry, **pLast;
        node2->exprList.exprs = entry = (ExprListEntry *)Alloc("NOD", sizeof(ExprListEntry));
        entry->expr = node;
        entry->next = NULL;
        pLast = &entry->next;
        do {
            entry = (ExprListEntry *)Alloc("EXP", sizeof(ExprListEntry));
            entry->expr = ParseExpr2(c);
            entry->next = NULL;
            *pLast = entry;
            pLast = &entry->next;
        } while ((tkn = GetToken(c)) == T_OR);
        node = node2;
    }
    SaveToken(c, tkn);
    return node;
}

/* ParseExpr2 - handle the AND operator */
static ParseTreeNode *ParseExpr2(ParseContext *c)
{
    ParseTreeNode *node;
    Token tkn;
    node = ParseExpr3(c);
    if ((tkn = GetToken(c)) == T_AND) {
        ParseTreeNode *node2 = NewParseTreeNode(NodeTypeConjunction);
        ExprListEntry *entry, **pLast;
        node2->exprList.exprs = entry = (ExprListEntry *)Alloc("EXP", sizeof(ExprListEntry));
        entry->expr = node;
        entry->next = NULL;
        pLast = &entry->next;
        do {
            entry = (ExprListEntry *)Alloc("EXP", sizeof(ExprListEntry));
            entry->expr = ParseExpr2(c);
            entry->next = NULL;
            *pLast = entry;
            pLast = &entry->next;
        } while ((tkn = GetToken(c)) == T_AND);
        node = node2;
    }
    SaveToken(c, tkn);
    return node;
}

/* ParseExpr3 - handle the BXOR operator */
static ParseTreeNode *ParseExpr3(ParseContext *c)
{
    ParseTreeNode *node, *node2;
    Token tkn;
    node = ParseExpr4(c);
    while ((tkn = GetToken(c)) == '^') {
        node2 = ParseExpr4(c);
        node = MakeBinaryOpNode(OP_BXOR, node, node2, NodeType(c, node));
    }
    SaveToken(c,tkn);
    return node;
}

/* ParseExpr4 - handle the BOR operator */
static ParseTreeNode *ParseExpr4(ParseContext *c)
{
    ParseTreeNode *node, *node2;
    Token tkn;
    node = ParseExpr5(c);
    while ((tkn = GetToken(c)) == '|') {
        node2 = ParseExpr5(c);
        node = MakeBinaryOpNode(OP_BOR, node, node2, NodeType(c, node));
    }
    SaveToken(c,tkn);
    return node;
}

/* ParseExpr5 - handle the BAND operator */
static ParseTreeNode *ParseExpr5(ParseContext *c)
{
    ParseTreeNode *node, *node2;
    Token tkn;
    node = ParseExpr6(c);
    while ((tkn = GetToken(c)) == '&') {
        node2 = ParseExpr6(c);
        node = MakeBinaryOpNode(OP_BAND, node, node2, NodeType(c, node));
    }
    SaveToken(c,tkn);
    return node;
}

/* ParseExpr6 - handle the '=' and '<>' operators */
static ParseTreeNode *ParseExpr6(ParseContext *c)
{
    ParseTreeNode *expr, *expr2;
    Token tkn;
    expr = ParseExpr7(c);
    while ((tkn = GetToken(c)) == '=' || tkn == T_NE) {
        int type, type2, op;
        expr2 = ParseExpr7(c);
        type = NodeType(c, expr)->id;
        type2 = NodeType(c, expr2)->id;
        if (type != TYPE_INTEGER && type != TYPE_STRING)
            ParseError(c, "expecting numbers or strings");
        else if (type != type2)
            ParseError(c, "types do not match");
        switch (tkn) {
        case '=':
            op = type == TYPE_INTEGER ? OP_EQ : OP_EQS;
            break;
        case T_NE:
            op = type == TYPE_INTEGER ? OP_NE : OP_NES;
            break;
        }
        expr = MakeBinaryOpNode(op, expr, expr2, &c->integerType);
    }
    SaveToken(c,tkn);
    return expr;
}

/* ParseExpr7 - handle the '<', '<=', '>=' and '>' operators */
static ParseTreeNode *ParseExpr7(ParseContext *c)
{
    ParseTreeNode *expr, *expr2;
    Token tkn;
    expr = ParseExpr8(c);
    while ((tkn = GetToken(c)) == '<' || tkn == T_LE || tkn == T_GE || tkn == '>') {
        int type, type2, op;
        expr2 = ParseExpr8(c);
        type = NodeType(c, expr)->id;
        type2 = NodeType(c, expr2)->id;
        if (type != TYPE_INTEGER && type != TYPE_STRING)
            ParseError(c, "expecting numbers or strings");
        else if (type != type2)
            ParseError(c, "types do not match");
        switch (tkn) {
        case '<':
            op = (type == TYPE_INTEGER ? OP_LT : OP_LTS);
            break;
        case T_LE:
            op = (type == TYPE_INTEGER ? OP_LE : OP_LES);
            break;
        case T_GE:
            op = (type == TYPE_INTEGER ? OP_GE : OP_GES);
            break;
        case '>':
            op = (type == TYPE_INTEGER ? OP_GT : OP_GTS);
            break;
        }
        expr = MakeBinaryOpNode(op, expr, expr2, &c->integerType);
    }
    SaveToken(c,tkn);
    return expr;
}

/* ParseExpr8 - handle the '<<' and '>>' operators */
static ParseTreeNode *ParseExpr8(ParseContext *c)
{
    ParseTreeNode *node, *node2;
    Token tkn;
    node = ParseExpr9(c);
    while ((tkn = GetToken(c)) == T_SHL || tkn == T_SHR) {
        int op;
        switch (tkn) {
        case T_SHL:
            op = OP_SHL;
            break;
        case T_SHR:
            op = OP_SHR;
            break;
        }
        node2 = ParseExpr9(c);
        node = MakeBinaryOpNode(op, node, node2, NodeType(c, node));
    }
    SaveToken(c,tkn);
    return node;
}

/* ParseExpr9 - handle the '+' and '-' operators */
static ParseTreeNode *ParseExpr9(ParseContext *c)
{
    ParseTreeNode *expr, *expr2;
    Token tkn;
    expr = ParseExpr10(c);
    while ((tkn = GetToken(c)) == '+' || tkn == '-') {
        expr2 = ParseExpr10(c);
        if (IsIntegerLit(expr) && IsIntegerLit(expr2)) {
            switch (tkn) {
            case '+':
                expr->integerLit.value = expr->integerLit.value + expr2->integerLit.value;
                break;
            case '-':
                expr->integerLit.value = expr->integerLit.value - expr2->integerLit.value;
                break;
            }
            FreeParseTree(expr2);
        }
        else {
            int type, type2, op;
            type = NodeType(c, expr)->id;
            type2 = NodeType(c, expr2)->id;
            if (type != TYPE_INTEGER && type != TYPE_STRING)
                ParseError(c, "expecting numbers or strings");
            else if (type != type2)
            ParseError(c, "types do not match");
            switch (tkn) {
            case '+':
                op = type == TYPE_INTEGER ? OP_ADD : OP_CAT;
                break;
            case '-':
                if (type == TYPE_STRING)
                    ParseError(c, "expecting a number");
                op = OP_SUB;
                break;
            }
            expr = MakeBinaryOpNode(op, expr, expr2, NodeType(c, expr));
        }
    }
    SaveToken(c, tkn);
    return expr;
}

/* ParseExpr10 - handle the '*', '/' and MOD operators */
static ParseTreeNode *ParseExpr10(ParseContext *c)
{
    ParseTreeNode *node, *node2;
    Token tkn;
    node = ParseExpr11(c);
    while ((tkn = GetToken(c)) == '*' || tkn == '/' || tkn == T_MOD) {
        node2 = ParseExpr11(c);
        if (IsIntegerLit(node) && IsIntegerLit(node2)) {
            switch (tkn) {
            case '*':
                node->integerLit.value = node->integerLit.value * node2->integerLit.value;
                break;
            case '/':
                if (node2->integerLit.value == 0)
                    ParseError(c, "division by zero in constant expression");
                node->integerLit.value = node->integerLit.value / node2->integerLit.value;
                break;
            case T_MOD:
                if (node2->integerLit.value == 0)
                    ParseError(c, "division by zero in constant expression");
                node->integerLit.value = node->integerLit.value % node2->integerLit.value;
                break;
            }
            FreeParseTree(node2);
        }
        else {
            int op;
            switch (tkn) {
            case '*':
                op = OP_MUL;
                break;
            case '/':
                op = OP_DIV;
                break;
            case T_MOD:
                op = OP_REM;
                break;
            }
            node = MakeBinaryOpNode(op, node, node2, NodeType(c, node));
        }
    }
    SaveToken(c, tkn);
    return node;
}

/* ParseExpr11 - handle unary operators */
static ParseTreeNode *ParseExpr11(ParseContext *c)
{
    ParseTreeNode *node;
    Token tkn;
    switch (tkn = GetToken(c)) {
    case '+':
        node = ParsePrimary(c);
        break;
    case '-':
        node = ParsePrimary(c);
        if (IsIntegerLit(node))
            node->integerLit.value = -node->integerLit.value;
        else
            node = MakeUnaryOpNode(OP_NEG, node, NodeType(c, node));
        break;
    case T_NOT:
        node = ParsePrimary(c);
        if (IsIntegerLit(node))
            node->integerLit.value = !node->integerLit.value;
        else
            node = MakeUnaryOpNode(OP_NOT, node, NodeType(c, node));
        break;
    case '~':
        node = ParsePrimary(c);
        if (IsIntegerLit(node))
            node->integerLit.value = ~node->integerLit.value;
        else
            node = MakeUnaryOpNode(OP_BNOT, node, NodeType(c, node));
        break;
    default:
        SaveToken(c,tkn);
        node = ParsePrimary(c);
        break;
    }
    return node;
}

/* ParsePrimary - parse function calls and array references */
static ParseTreeNode *ParsePrimary(ParseContext *c)
{
    Type *type;
    ParseTreeNode *node = ParseSimplePrimary(c);
    type = NodeType(c, node);
    if (type->id == TYPE_FUNCTION) {
        if (!type->FunctionInfo.returnType)
            ParseError(c, "attempt to call a subroutine as a function");
        node = ParseCall(c, node);
    }
    else
        node = ParsePossibleArrayReference(c, node);
    return node;
}

/* ParsePossibleArrayReference - parse a possible array reference after a primary expression */
ParseTreeNode *ParsePossibleArrayReference(ParseContext *c, ParseTreeNode *node)
{
    Token tkn;
    if ((tkn = GetToken(c)) == '(') {
        if (NodeType(c, node)->id == TYPE_ARRAY)
            node = ParseArrayReference(c, node);
        else
            ParseError(c, "expecting an array");
    }
    else
        SaveToken(c, tkn);
    return node;
}

/* ParseArrayReference - parse an array reference */
static ParseTreeNode *ParseArrayReference(ParseContext *c, ParseTreeNode *arrayNode)
{
    ParseTreeNode *node = NewParseTreeNode(NodeTypeArrayRef);
    Token tkn;

    /* setup the array reference */
    node->arrayRef.array = arrayNode;

    /* get the first index expression */
    node->arrayRef.index1 = ParseExpr(c);
    
    /* check for a second index */
    if ((tkn = GetToken(c)) == ',') {
        node->arrayRef.index2 = ParseExpr(c);
        tkn = GetToken(c);
    }
    
    /* check for the close paren */
    Require(c, tkn, ')');
    return node;
}

/* ParseSubroutineCall - parse a call to a subroutine */
ParseTreeNode *ParseSubroutineCall(ParseContext *c, ParseTreeNode *functionNode)
{
    Type *type;
    /* make sure it's a function we're calling */
    type = NodeType(c, functionNode);
    if (type->id != TYPE_FUNCTION)
        ParseError(c, "Attempt to call a non-subroutine");
    else if (type->FunctionInfo.returnType)
        ParseError(c, "attempt to call a function as a subroutine");

    /* parse the call */
    return ParseCall(c, functionNode);
}

/* ParseCall - parse a function or subroutine call */
static ParseTreeNode *ParseCall(ParseContext *c, ParseTreeNode *functionNode)
{
    ParseTreeNode *node = NewParseTreeNode(NodeTypeFunctionCall);
    Symbol *arg;
    Type *type;
    Token tkn;

    /* intialize the function call node */
    node->functionCall.fcn = functionNode;
    
    /* get the list of argument types */
    type = NodeType(c, functionNode);
    arg = type->FunctionInfo.arguments.head;
    if (type->FunctionInfo.returnType)
        arg = arg->next;

    /* parse the optional argument list */
    if ((tkn = GetToken(c)) == '(') {
        ExprListEntry **pLast = &node->functionCall.args;

        /* compile each argument expression */
        if ((tkn = GetToken(c)) != ')') {
            SaveToken(c, tkn);
            do {
                ExprListEntry *actual;

                /* allocate an actual argument structure */
                actual = (ExprListEntry *)Alloc("EXP", sizeof(ExprListEntry));
                actual->expr = ParseExpr(c);
                actual->next = NULL;
                *pLast = actual;
                pLast = &actual->next;

                /* check the argument count and type */
                if (!arg)
                    ParseError(c, "too many arguments");
                else if (!CompareTypes(NodeType(c, actual->expr), arg->type))
                    ParseError(c, "wrong argument type");

                /* move ahead to the next argument */
                ++node->functionCall.argc;
                arg = arg->next;
            } while ((tkn = GetToken(c)) == ',');
            Require(c, tkn, ')');
        }
    }

    /* make sure there werent' too many arguments specified */
    if (arg)
        ParseError(c, "too few arguments");

    /* return the function call node */
    return node;
}

/* ParseSimplePrimary - parse a primary expression */
static ParseTreeNode *ParseSimplePrimary(ParseContext *c)
{
    ParseTreeNode *node;
    switch (GetToken(c)) {
    case '(':
        node = ParseExpr(c);
        FRequire(c,')');
        break;
    case T_NUMBER:
        node = NewParseTreeNode(NodeTypeIntegerLit);
        node->integerLit.value = c->value;
        break;
    case T_STRING:
        node = NewParseTreeNode(NodeTypeStringLit);
        node->stringLit.string = AddString(c, c->token);
        break;
    case T_IDENTIFIER:
        node = GetSymbolRef(c, c->token);
        break;
    default:
        ParseError(c, "Expecting a primary expression");
        break;
    }
    return node;
}

/* GetSymbolRef - setup a symbol reference */
ParseTreeNode *GetSymbolRef(ParseContext *c, char *name)
{
    ParseTreeNode *node = NewParseTreeNode(NodeTypeSymbolRef);
    Symbol *symbol;

    /* handle local variables within a function or subroutine */
    if (c->function && (symbol = FindSymbol(&c->locals, name)) != NULL) {
        if (symbol->storageClass == SC_CONSTANT) {
            switch (symbol->type->id) {
            case TYPE_INTEGER:
                node->nodeType = NodeTypeIntegerLit;
                node->integerLit.value = symbol->value;
                break;
            default:
                ParseError(c, "unknown constant type");
                break;
            }
        }
        else {
            node->symbolRef.symbol = symbol;
            node->symbolRef.fcn = code_local;
            node->symbolRef.offset = symbol->offset;
        }
    }

    /* handle function or subroutine arguments or the return value symbol */
    else if (c->function && (symbol = FindSymbol(&c->functionType->FunctionInfo.arguments, name)) != NULL) {
        node->symbolRef.symbol = symbol;
        node->symbolRef.fcn = code_argument;
        node->symbolRef.offset = symbol->offset - c->functionType->FunctionInfo.argumentCount;
    }

    /* handle global symbols */
    else if ((symbol = FindSymbol(&c->globals, c->token)) != NULL) {
        if (symbol->storageClass == SC_CONSTANT) {
            switch (symbol->type->id) {
            case TYPE_INTEGER:
                node->nodeType = NodeTypeIntegerLit;
                node->integerLit.value = symbol->value;
                break;
            case TYPE_FUNCTION:
                node->nodeType = NodeTypeFunctionLit;
                node->functionLit.type = symbol->type;
                node->functionLit.offset = symbol->value;
                break;
            default:
                ParseError(c, "unknown constant type");
                break;
            }
        }
        else {
            node->symbolRef.symbol = symbol;
            node->symbolRef.fcn = code_global;
            node->symbolRef.offset = symbol->offset;
        }
    }

    /* handle undefined symbols */
    else {

        /* if inside a function or subroutine definition, add the symbol as a local */
        if (c->function) {
            if (!(symbol = AddSymbol(&c->locals, name, SC_VARIABLE, DefaultType(c, name), c->localCount++ + F_SIZE + 1, 0)))
                Fatal("insufficient memory");
            node->symbolRef.symbol = symbol;
            node->symbolRef.fcn = code_local;
            node->symbolRef.offset = symbol->offset;
        }

        /* otherwise, add it as a global */
        else {
            if (!(symbol = AddSymbol(&c->globals, name, SC_VARIABLE, DefaultType(c, name), c->globalCount++, 0)))
                Fatal("insufficient memory");
            node->symbolRef.symbol = symbol;
            node->symbolRef.fcn = code_global;
            node->symbolRef.offset = symbol->offset;
        }
    }

    /* return the symbol reference node */
    return node;
}

/* MakeUnaryOpNode - allocate a unary operation parse tree node */
static ParseTreeNode *MakeUnaryOpNode(int op, ParseTreeNode *expr, Type *resultType)
{
    ParseTreeNode *node = NewParseTreeNode(NodeTypeUnaryOp);
    node->unaryOp.op = op;
    node->unaryOp.expr = expr;
    node->unaryOp.resultType = resultType;
    return node;
}

/* MakeBinaryOpNode - allocate a binary operation parse tree node */
static ParseTreeNode *MakeBinaryOpNode(int op, ParseTreeNode *left, ParseTreeNode *right, Type *resultType)
{
    ParseTreeNode *node = NewParseTreeNode(NodeTypeBinaryOp);
    node->binaryOp.op = op;
    node->binaryOp.left = left;
    node->binaryOp.right = right;
    node->binaryOp.resultType = resultType;
    return node;
}

/* NewParseTreeNode - allocate a new parse tree node */
static ParseTreeNode *NewParseTreeNode(int type)
{
    ParseTreeNode *node = (ParseTreeNode *)Alloc("NOD", sizeof(ParseTreeNode));
    memset(node, 0, sizeof(ParseTreeNode));
    node->nodeType = type;
    return node;
}

/* FreeParseTree - free a parse tree */
void FreeParseTree(ParseTreeNode *expr)
{
    if (expr) {
        ExprListEntry *entry, *next;
        switch (expr->nodeType) {
        case NodeTypeUnaryOp:
            FreeParseTree(expr->unaryOp.expr);
            break;
        case NodeTypeBinaryOp:
            FreeParseTree(expr->binaryOp.left);
            FreeParseTree(expr->binaryOp.right);
            break;
        case NodeTypeArrayRef:
            FreeParseTree(expr->arrayRef.array);
            FreeParseTree(expr->arrayRef.index1);
            FreeParseTree(expr->arrayRef.index2);
            break;
        case NodeTypeFunctionCall:
            FreeParseTree(expr->functionCall.fcn);
            for (entry = expr->functionCall.args; entry != NULL; entry = next) {
                next = entry->next;
                FreeParseTree(entry->expr);
                Free(entry);
            }
            break;
        case NodeTypeDisjunction:
        case NodeTypeConjunction:
            for (entry = expr->exprList.exprs; entry != NULL; entry = next) {
                next = entry->next;
                FreeParseTree(entry->expr);
                Free(entry);
            }
            break;
        }
        Free(expr);
    }
}

/* NodeSrcType - get the type of a parse tree node */
Type *NodeSrcType(ParseContext *c, ParseTreeNode *expr)
{
    Type *type;
    switch (expr->nodeType) {
    case NodeTypeSymbolRef:
        type = expr->symbolRef.symbol->type;
        break;
    case NodeTypeStringLit:
        type = &c->stringType;
        break;
    case NodeTypeIntegerLit:
        type = &c->integerType;
        break;
    case NodeTypeFunctionLit:
        type = expr->functionLit.type;
        break;
    case NodeTypeUnaryOp:
        type = expr->unaryOp.resultType;
        break;
    case NodeTypeBinaryOp:
        type = expr->binaryOp.resultType;
        break;
    case NodeTypeArrayRef:
        type = NodeType(c, expr->arrayRef.array)->ArrayInfo.elementType;
        break;
    case NodeTypeFunctionCall:
        type = NodeType(c, expr->functionCall.fcn)->FunctionInfo.returnType;
        break;
    case NodeTypeDisjunction:
    case NodeTypeConjunction:
        type = &c->integerType;
        break;
    }
    return type;
}

/* NodeType - get the type of a parse tree node */
Type *NodeType(ParseContext *c, ParseTreeNode *expr)
{
    Type *type = NodeSrcType(c, expr);
    return type == &c->byteType ? &c->integerType : type;
}

/* IsIntegerLit - check to see if a node is an integer literal */
int IsIntegerLit(ParseTreeNode *node)
{
    return node->nodeType == NodeTypeIntegerLit;
}
