/* db_statement.c - statement parser
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdlib.h>
#include <string.h>
#include "db_compiler.h"

/* statement handler prototypes */
static void ParseOption(ParseContext *c);
static void ParseDef(ParseContext *c);
static void ParseEndDef(ParseContext *c);
static void ParseSub(ParseContext *c);
static void ParseEndSub(ParseContext *c);
static void ParseType(ParseContext *c);
static void ParseDim(ParseContext *c);
static Type *ParseVariableDecl(ParseContext *c, char *name);
static void ParseImpliedLetOrSubCall(ParseContext *c);
static void ParseLet(ParseContext *c);
static void ParseCall(ParseContext *c);
static void ParseIf(ParseContext *c);
static void ParseElse(ParseContext *c);
static void ParseElseIf(ParseContext *c);
static void ParseEndIf(ParseContext *c);
static void ParseSelect(ParseContext *c);
static void ParseCase(ParseContext *c);
static void ParseEndSelect(ParseContext *c);
static void ParseEnd(ParseContext *c);
static void ParseConstantDef(ParseContext *c, char *name);
static void ParseDefOrSub(ParseContext *c, char *name, int isFunction);
static void ParseEndDefOrSub(ParseContext *c);
static void ParseFor(ParseContext *c);
static void ParseNext(ParseContext *c);
static void ParseDo(ParseContext *c);
static void ParseDoWhile(ParseContext *c);
static void ParseDoUntil(ParseContext *c);
static void ParseLoop(ParseContext *c);
static void ParseLoopWhile(ParseContext *c);
static void ParseLoopUntil(ParseContext *c);
static void ParseStop(ParseContext *c);
static void ParsePrint(ParseContext *c);
static void ParseWrite(ParseContext *c);
static void ParseDisplay(ParseContext *c);
static void ParseInput(ParseContext *c);
static void ParseOpen(ParseContext *c);
static void ParseClose(ParseContext *c);
static void ParseGoto(ParseContext *c);

/* prototypes */
static void DefineLabel(ParseContext *c, char *name, int offset);
static int ReferenceLabel(ParseContext *c, char *name, int offset);
static void CheckLabels(ParseContext *c);
static BlockType CheckBlock(ParseContext *c);
static void PushBlock(ParseContext *c);
static void PopBlock(ParseContext *c);

/* ParseStatement - parse a statement */
void ParseStatement(ParseContext *c)
{
    Token tkn;
    
    /* dispatch on the statement keyword */
    switch (tkn = GetToken(c)) {
    case T_REM:
        /* just a comment so ignore the rest of the line */
        break;
    case T_OPTION:
        ParseOption(c);
        break;
    case T_DEF:
        ParseDef(c);
        break;
    case T_END_DEF:
        ParseEndDef(c);
        break;
    case T_SUB:
        ParseSub(c);
        break;
    case T_END_SUB:
        ParseEndSub(c);
        break;
    case T_TYPE:
        ParseType(c);
        break;
    case T_DIM:
        ParseDim(c);
        break;
    case T_LET:
        ParseLet(c);
        break;
    case T_CALL:
        ParseCall(c);
        break;
    case T_IF:
        ParseIf(c);
        break;
    case T_ELSE:
        ParseElse(c);
        break;
    case T_ELSE_IF:
        ParseElseIf(c);
        break;
    case T_END_IF:
        ParseEndIf(c);
        break;
    case T_SELECT:
        ParseSelect(c);
        break;
    case T_CASE:
        ParseCase(c);
        break;
    case T_END_SELECT:
        ParseEndSelect(c);
        break;
    case T_END:
        ParseEnd(c);
        break;
    case T_FOR:
        ParseFor(c);
        break;
    case T_NEXT:
        ParseNext(c);
        break;
    case T_DO:
        ParseDo(c);
        break;
    case T_DO_WHILE:
        ParseDoWhile(c);
        break;
    case T_DO_UNTIL:
        ParseDoUntil(c);
        break;
    case T_LOOP:
        ParseLoop(c);
        break;
    case T_LOOP_WHILE:
        ParseLoopWhile(c);
        break;
    case T_LOOP_UNTIL:
        ParseLoopUntil(c);
        break;
    case T_STOP:
        ParseStop(c);
        break;
    case T_PRINT:
        ParsePrint(c);
        break;
    case T_WRITE:
        ParseWrite(c);
        break;
    case T_DISPLAY:
        ParseDisplay(c);
        break;
    case T_INPUT:
        ParseInput(c);
        break;
    case T_OPEN:
        ParseOpen(c);
        break;
    case T_CLOSE:
        ParseClose(c);
        break;
    case T_GOTO:
        ParseGoto(c);
        break;
    case T_IDENTIFIER:
        if (SkipSpaces(c) == ':') {
            DefineLabel(c, c->token, codeaddr(c));
            break;
        }
        UngetC(c);
        ParseImpliedLetOrSubCall(c);
        break;
    default:
        ParseError(c, "unknown statement type");
    }
}

/* ParseOption - parse the 'OPTION' statement */
static void ParseOption(ParseContext *c)
{
    char option[MAXTOKEN];

    FRequire(c, T_IDENTIFIER);
    strcpy(option, c->token);
    FRequire(c, '=');
    FRequire(c, T_STRING);

    /* handle the 'target' option */
    if (strcasecmp(option, "target") == 0)
        SelectRuntime(c, c->token);

    /* handle the 'heap' option */
    else if (strcasecmp(option, "heap") == 0)
        SetHeapSize(c, c->token);

    /* unknown option */
    else
        ParseError(c, "unknown option: %s", option);
}

/* ParseDef - parse the 'DEF' statement */
static void ParseDef(ParseContext *c)
{
    char name[MAXTOKEN];
    Token tkn;

    /* get the name being defined */
    FRequire(c, T_IDENTIFIER);
    strcpy(name, c->token);

    /* check for a constant definition */
    if ((tkn = GetToken(c)) == '=')
        ParseConstantDef(c, name);
    
    /* otherwise, assume a function definition */
    else {
        SaveToken(c, tkn);
        ParseDefOrSub(c, name, VMTRUE);
    }
}

/* ParseEndDef - parse the 'END DEF' statement */
static void ParseEndDef(ParseContext *c)
{
    ParseEndDefOrSub(c);
}

/* ParseSub - parse the 'SUB' statement */
static void ParseSub(ParseContext *c)
{
    char name[MAXTOKEN];
    FRequire(c, T_IDENTIFIER);
    strcpy(name, c->token);
    ParseDefOrSub(c, name, VMFALSE);
}

/* ParseEndSub - parse the 'END SUB' statement */
static void ParseEndSub(ParseContext *c)
{
    ParseEndDefOrSub(c);
}

/* ParseConstantDef - parse a 'DEF <name> =' statement */
static void ParseConstantDef(ParseContext *c, char *name)
{
    ParseTreeNode *expr;
    Symbol *sym;

    /* get the constant value */
    expr = ParseExpr(c);

    /* make sure it's a constant */
    if (!IsIntegerLit(expr))
        ParseError(c, "expecting a constant expression");

    /* if inside a function or subroutine definition, add the symbol as a local */
    if (c->function) {
        if (!(sym = AddSymbol(&c->locals, name, SC_CONSTANT, &c->integerType, 0, expr->integerLit.value)))
            Fatal("insufficient memory");
    }

    /* otherwise, add it as a global */
    else {
        if (!(sym = AddSymbol(&c->globals, name, SC_CONSTANT, &c->integerType, 0, expr->integerLit.value)))
            Fatal("insufficient memory");
    }

    /* free the parse tree */
    FreeParseTree(expr);

    FRequire(c, T_EOL);
}

/* ParseDefOrSub - parse a 'DEF' or 'SUB' statement */
static void ParseDefOrSub(ParseContext *c, char *name, int isFunction)
{
    Token tkn;

    /* all function and subroutines must precede the main code */
    if (c->cptr > c->codeBuf)
        ParseError(c, "functions and subroutines must precede the main code");

    /* don't allow nested functions or subroutines (for now anyway) */
    if (c->function)
        ParseError(c, "nested functions not supported");

    /* create a type for the function */
    c->functionType = NewType(TYPE_FUNCTION);
    InitSymbolTable(&c->functionType->FunctionInfo.arguments);
    c->functionType->FunctionInfo.argumentCount = 0;

    /* initialize the local symbol table */
    InitSymbolTable(&c->locals);
    c->localCount = 0;

    /* make sure the function starts at an even address */
    if (c->codeOff & 1) {
        putc('\0', c->ofp);
        ++c->codeOff;
    }
    
    /* set the return type for functions */
    if (isFunction) {
        if (!AddSymbol(&c->functionType->FunctionInfo.arguments, name, SC_VARIABLE, DefaultType(c, name), 0, 0))
            Fatal("insufficient memory");
        c->functionType->FunctionInfo.returnType = &c->integerType;
    }

    /* no return type for subroutines */
    else
        c->functionType->FunctionInfo.returnType = NULL;

    /* get the argument list */
    if ((tkn = GetToken(c)) == '(') {
        if ((tkn = GetToken(c)) != ')') {
            int offset = 1;
            SaveToken(c, tkn);
            do {
                char name[MAXTOKEN];
                Type *type;
                type = ParseVariableDecl(c, name);
                if (!AddSymbol(&c->functionType->FunctionInfo.arguments, name, SC_VARIABLE, type, offset, 0))
                    Fatal("insufficient memory");
                ++c->functionType->FunctionInfo.argumentCount;
                ++offset;
            } while ((tkn = GetToken(c)) == ',');
        }
        Require(c, tkn, ')');
    }
    else
        SaveToken(c, tkn);

    /* enter the function name in the global symbol table */
    if (!(c->function = AddSymbol(&c->globals, name, SC_CONSTANT, c->functionType, 0, c->codeOff)))
        Fatal("insufficient memory");

    /* create the function prolog */
    putcbyte(c, OP_RESERVE);
    putcbyte(c, 0);
    
    FRequire(c, T_EOL);
}

/* ParseEndDefOrSub - parse an 'END DEF' or 'END SUB' statement */
static void ParseEndDefOrSub(ParseContext *c)
{
    if (!c->functionType)
        ParseError(c, "not in a function definition");
    else {
        Symbol *sym = c->functionType->FunctionInfo.arguments.head;
        while (sym) {
            if (sym->offset > 0 && sym->type->id == TYPE_STRING) {
                int off = sym->offset - c->functionType->FunctionInfo.argumentCount;
                putcbyte(c, OP_LDECREF);
                putcbyte(c, off);
            }
            sym = sym->next;
        }
        sym = c->locals.head;
        while (sym) {
            if (sym->type->id == TYPE_STRING) {
                putcbyte(c, OP_LDECREF);
                putcbyte(c, sym->offset);
            }
            sym = sym->next;
        }
        if (c->functionType->FunctionInfo.returnType)
            putcbyte(c, OP_FRET);
        else
            putcbyte(c, OP_SRET);
        c->codeBuf[1] = c->localCount;
        CheckLabels(c);
        WriteFunction(c, c->function->name);
        if (c->debugMode) {
            if (c->functionType->FunctionInfo.arguments.head) {
                printf("arguments:\n");
                DumpSymbols(&c->functionType->FunctionInfo.arguments, VMFALSE);
            }
            if (c->locals.head) {
                printf("locals:\n");
                DumpSymbols(&c->locals, VMFALSE);
            }
            printf("\n");
        }
        c->function = NULL;
        c->functionType = NULL;
        FreeSymbols(c, &c->locals);
    }
}

/* ParseType - parse the 'TYPE' statement */
static void ParseType(ParseContext *c)
{
    /*
        type <name> as <type>
        type <name> as <type> ( [ <size> ] [ , [ <size> ] ] )
        type <name> as def [ (<args>) ]
        type <name> as sub [ (<args>) ]
        type <name>
          <field-name> [ as <type> ]
          ...
        end type

        <args>:
            <arg>
            <args> , <arg>

        <arg>:
            <name> [ as <type> ]
    */
}

/* ParseDim - parse the 'DIM' statement */
static void ParseDim(ParseContext *c)
{
    int valuePrefix, valueSize;
    char name[MAXTOKEN];
    Symbol *sym;
    Type *type;
    Token tkn;

    /* parse variable declarations */
    do {

        /* parse a variable declaration */
        type = ParseVariableDecl(c, name);

        /* get the size of the value */
        valueSize = ValueSize(type, &valuePrefix);

        /* add to the local symbol table if inside a function definition */
        if (c->function) {
            int offset = c->localCount + F_SIZE + 1 + valuePrefix;
            if (type->id == TYPE_ARRAY) {
                putcbyte(c, OP_LIT);
                putcword(c, type->ArrayInfo.dim1);
                putcbyte(c, OP_LSET);
                putcbyte(c, offset - 1);
                putcbyte(c, OP_LIT);
                putcword(c, type->ArrayInfo.dim1 * type->ArrayInfo.dim2);
                putcbyte(c, OP_LSET);
                putcbyte(c, offset - 2);
            }
            if (!(sym = AddSymbol(&c->locals, name, SC_VARIABLE, type, offset, 0)))
                Fatal("insufficient memory");
            c->localCount += valueSize;
        }
        
        /* otherwise, add to the global symbol table */
        else {
            if (!(sym = AddSymbol(&c->globals, name, SC_VARIABLE, type, c->globalCount + valuePrefix, 0)))
                Fatal("insufficient memory");
            c->globalCount += valueSize;
        }

    } while ((tkn = GetToken(c)) == ',');

    Require(c, tkn, T_EOL);
}

/* ParseVariableDecl - parse a variable declaration */
static Type *ParseVariableDecl(ParseContext *c, char *name)
{
    Type *type, **pBaseType;
    Token tkn;

    /* parse the variable name */
    FRequire(c, T_IDENTIFIER);
    strcpy(name, c->token);
    
    /* determine the default type based on the variable name */
    type = DefaultType(c, name);
    pBaseType = &type;

    /* handle arrays */
    if ((tkn = GetToken(c)) == '(') {
        Type *arrayType;
        int dim1, dim2;

        /* check for an array with unspecified dimensions */
        if ((tkn = GetToken(c)) == ')')
            dim1 = dim2 = 0;

        /* otherwise, parse the dimensions */
        else {
            ParseTreeNode *expr;

            /* put back the token */
            SaveToken(c, tkn);

            /* get the array size */
            expr = ParseExpr(c);

            /* make sure it's a constant */
            if (!IsIntegerLit(expr) || expr->integerLit.value <= 0)
                ParseError(c, "expecting a positive constant expression");
            dim1 = expr->integerLit.value;
            FreeParseTree(expr);

            /* check for a second dimension */
            if ((tkn = GetToken(c)) == ',') {
                expr = ParseExpr(c);
                if (!IsIntegerLit(expr) || expr->integerLit.value <= 0)
                    ParseError(c, "expecting a positive constant expression");
                dim2 = expr->integerLit.value;
                FreeParseTree(expr);
                tkn = GetToken(c);
            }

            /* no second dimension */
            else
                dim2 = 1;

            /* only two dimensions allowed for now */
            Require(c, tkn, ')');
        }

        /* create a type for the function */
        arrayType = NewType(TYPE_ARRAY);
        arrayType->ArrayInfo.elementType = type;
        pBaseType = &arrayType->ArrayInfo.elementType;
        arrayType->ArrayInfo.dim1 = dim1;
        arrayType->ArrayInfo.dim2 = dim2;
        type = arrayType;

        tkn = GetToken(c);
    }

    /* check for a type specification */
    if (tkn == T_AS) {
        Type *declaredType;
        FRequire(c, T_IDENTIFIER);
        if (!(declaredType = FindType(&c->types, c->token)))
            ParseError(c, "unknown type: %s", c->token);
        *pBaseType = declaredType;
    }

    /* just use the default type */
    else
        SaveToken(c, tkn);

    /* return the type */
    return type;
}

/* ParseImpliedLetOrSubCall - parse an implied statement or a subroutine call */
static void ParseImpliedLetOrSubCall(ParseContext *c)
{
    ParseTreeNode *expr;

    /* get the variable or subroutine value */
    expr = GetSymbolRef(c, c->token);

    /* check for a subroutine call */
    if (NodeType(c, expr)->id == TYPE_FUNCTION) {
        expr = ParseSubroutineCall(c, expr);
        code_rvalue(c, expr);
    }

    /* handle an implied let */
    else {
        PVAL pv;
        expr = ParsePossibleArrayReference(c, expr);
        code_lvalue(c, expr, &pv);
        FRequire(c, '=');
        ParseRValue(c);
        (*pv.fcn)(c, PV_STORE, &pv);
    }
    FreeParseTree(expr);
    FRequire(c, T_EOL);
}

/* ParseLet - parse the 'LET' statement */
static void ParseLet(ParseContext *c)
{
    ParseTreeNode *lvalue;
    PVAL pv;
    FRequire(c, T_IDENTIFIER);
    lvalue = GetSymbolRef(c, c->token);
    lvalue = ParsePossibleArrayReference(c, lvalue);
    code_lvalue(c, lvalue, &pv);
    FRequire(c, '=');
    ParseRValue(c);
    (*pv.fcn)(c, PV_STORE, &pv);
    FreeParseTree(lvalue);
    FRequire(c, T_EOL);
}

/* ParseCall - parse the 'CALL' statement */
static void ParseCall(ParseContext *c)
{
    ParseTreeNode *expr;
    FRequire(c, T_IDENTIFIER);
    expr = GetSymbolRef(c, c->token);
    expr = ParseSubroutineCall(c, expr);
    code_rvalue(c, expr);
    FreeParseTree(expr);
    FRequire(c, T_EOL);
}

/* ParseIf - parse the 'IF' statement */
static void ParseIf(ParseContext *c)
{
    Token tkn;
    ParseRValue(c);
    FRequire(c, T_THEN);
    PushBlock(c);
    c->bptr->type = BLOCK_IF;
    putcbyte(c, OP_BRF);
    c->bptr->IfBlock.nxt = putcword(c, 0);
    c->bptr->IfBlock.end = 0;
    if ((tkn = GetToken(c)) != T_EOL) {
        SaveToken(c, tkn);
        ParseStatement(c);
        fixupbranch(c, c->bptr->IfBlock.nxt, codeaddr(c));
        PopBlock(c);
    }
    else
        Require(c, tkn, T_EOL);
}

/* ParseElseIf - parse the 'ELSE IF' statement */
static void ParseElseIf(ParseContext *c)
{
    switch (CheckBlock(c)) {
    case BLOCK_IF:
        putcbyte(c, OP_BR);
        c->bptr->IfBlock.end = putcword(c, c->bptr->IfBlock.end);
        fixupbranch(c, c->bptr->IfBlock.nxt, codeaddr(c));
        c->bptr->IfBlock.nxt = 0;
        ParseRValue(c);
        FRequire(c, T_THEN);
        putcbyte(c, OP_BRF);
        c->bptr->IfBlock.nxt = putcword(c, 0);
        FRequire(c, T_EOL);
        break;
    default:
        ParseError(c, "ELSE IF without a matching IF");
        break;
    }
}

/* ParseElse - parse the 'ELSE' statement */
static void ParseElse(ParseContext *c)
{
    int end;
    switch (CheckBlock(c)) {
    case BLOCK_IF:
        putcbyte(c, OP_BR);
        end = putcword(c, c->bptr->IfBlock.end);
        fixupbranch(c, c->bptr->IfBlock.nxt, codeaddr(c));
        c->bptr->type = BLOCK_ELSE;
        c->bptr->ElseBlock.end = end;
        break;
    default:
        ParseError(c, "ELSE without a matching IF");
        break;
    }
    FRequire(c, T_EOL);
}

/* ParseEndIf - parse the 'END IF' statement */
static void ParseEndIf(ParseContext *c)
{
    switch (CheckBlock(c)) {
    case BLOCK_IF:
        fixupbranch(c, c->bptr->IfBlock.nxt, codeaddr(c));
        fixupbranch(c, c->bptr->IfBlock.end, codeaddr(c));
        PopBlock(c);
        break;
    case BLOCK_ELSE:
        fixupbranch(c, c->bptr->ElseBlock.end, codeaddr(c));
        PopBlock(c);
        break;
    default:
        ParseError(c, "ELSE without a matching IF");
        break;
    }
    FRequire(c, T_EOL);
}


/* ParseSelect - parse the 'SELECT' statement */
static void ParseSelect(ParseContext *c)
{
    ParseRValue(c);
    PushBlock(c);
    c->bptr->type = BLOCK_SELECT;
    c->bptr->SelectBlock.first = VMTRUE;
    c->bptr->SelectBlock.nxt = 0;
    c->bptr->SelectBlock.end = 0;
    FRequire(c, T_EOL);
}

/* ParseCase - parse the 'CASE' statement */
static void ParseCase(ParseContext *c)
{
    Token tkn;

    /* handle 'CASE ELSE' statement */
    if ((tkn = GetToken(c)) == T_ELSE) {
        int end;
        switch (CheckBlock(c)) {
        case BLOCK_SELECT:

            /* finish the previous case */
            if (c->bptr->SelectBlock.first) {
                c->bptr->SelectBlock.first = VMFALSE;
                end = 0;
            }
            else {
                putcbyte(c, OP_BR);
                end = putcword(c, c->bptr->SelectBlock.end);
            }

            /* start the else case */
            fixupbranch(c, c->bptr->SelectBlock.nxt, codeaddr(c));
            c->bptr->type = BLOCK_CASE_ELSE;
            c->bptr->CaseElseBlock.end = end;
            break;
        default:
            ParseError(c, "CASE ELSE without a matching SELECT");
            break;
        }
        FRequire(c, T_EOL);
    }

    /* handle 'CASE expr...' statement */
    else {
        int alt = 0;
        int body = 0;
        SaveToken(c, tkn);
        switch (CheckBlock(c)) {
        case BLOCK_SELECT:

            /* finish the previous case */
            if (c->bptr->SelectBlock.first)
                c->bptr->SelectBlock.first = VMFALSE;
            else {
                putcbyte(c, OP_BR);
                c->bptr->SelectBlock.end = putcword(c, c->bptr->SelectBlock.end);
            }
            
            fixupbranch(c, c->bptr->SelectBlock.nxt, codeaddr(c));
            c->bptr->SelectBlock.nxt = 0;

            /* handle a list of comma separated expressions or ranges */
            do {

                if (alt) {
                    fixupbranch(c, alt, codeaddr(c));
                    alt = 0;
                }

                putcbyte(c, OP_DUP);
                ParseRValue(c);

                /* handle an 'expr TO expr' range */
                if ((tkn = GetToken(c)) == T_TO) {

                    /* check the lower bound */
                    putcbyte(c, OP_GE);
                    putcbyte(c, OP_BRF);
                    alt = putcword(c, alt);

                    /* check the upper bound */
                    putcbyte(c, OP_DUP);
                    ParseRValue(c);
                    putcbyte(c, OP_LE);
                }

                /* handle a single expression */
                else {
                    SaveToken(c, tkn);
                    putcbyte(c, OP_EQ);
                }

                /* more expressions or ranges follow */
                if ((tkn = GetToken(c)) == ',') {
                    putcbyte(c, OP_BRT);
                    body = putcword(c, body);
                }

                /* last expression or range */
                else {
                    putcbyte(c, OP_BRF);
                    c->bptr->SelectBlock.nxt = putcword(c, c->bptr->SelectBlock.nxt);
                }
            } while (tkn == ',');
            c->bptr->SelectBlock.nxt = merge(c, c->bptr->SelectBlock.nxt, alt);
            fixupbranch(c, body, codeaddr(c));
            Require(c, tkn, T_EOL);
            break;
        default:
            ParseError(c, "CASE without a matching SELECT");
            break;
        }
    }
}

/* ParseEndSelect - parse the 'END SELECT' statement */
static void ParseEndSelect(ParseContext *c)
{
    switch (CheckBlock(c)) {
    case BLOCK_SELECT:
        fixupbranch(c, c->bptr->SelectBlock.nxt, codeaddr(c));
        fixupbranch(c, c->bptr->SelectBlock.end, codeaddr(c));
        PopBlock(c);
        break;
    case BLOCK_CASE_ELSE:
        fixupbranch(c, c->bptr->CaseElseBlock.end, codeaddr(c));
        PopBlock(c);
        break;
    default:
        ParseError(c, "END SELECT without a matching SELECT");
        break;
    }
    putcbyte(c, OP_DROP);
    FRequire(c, T_EOL);
}

/* ParseEnd - parse the 'END' statement */
static void ParseEnd(ParseContext *c)
{
    putcbyte(c, OP_HALT);
    FRequire(c, T_EOL);
}

/* ParseFor - parse the 'FOR' statement */
static void ParseFor(ParseContext *c)
{
    ParseTreeNode *var, *step;
    int test, body, inst;
    Token tkn;
    PVAL pv;

    PushBlock(c);
    c->bptr->type = BLOCK_FOR;

    /* get the control variable */
    FRequire(c, T_IDENTIFIER);
    var = GetSymbolRef(c, c->token);
    code_lvalue(c, var, &pv);
    FreeParseTree(var);
    FRequire(c, '=');

    /* parse the starting value expression */
    ParseRValue(c);

    /* parse the TO expression and generate the loop termination test */
    test = codeaddr(c);
    (*pv.fcn)(c, PV_STORE, &pv);
    (*pv.fcn)(c, PV_LOAD, &pv);
    FRequire(c, T_TO);
    ParseRValue(c);
    putcbyte(c, OP_LE);
    putcbyte(c, OP_BRT);
    body = putcword(c, 0);

    /* branch to the end if the termination test fails */
    putcbyte(c, OP_BR);
    c->bptr->ForBlock.end = putcword(c, 0);
    
    /* update the for variable after an iteration of the loop */
    c->bptr->ForBlock.nxt = codeaddr(c);
    (*pv.fcn)(c, PV_LOAD, &pv);

    /* get the STEP expression */
    if ((tkn = GetToken(c)) == T_STEP) {
        step = ParseExpr(c);
        code_rvalue(c, step);
        FreeParseTree(step);
        tkn = GetToken(c);
    }

    /* no step so default to one */
    else {
        putcbyte(c, OP_LIT);
        putcword(c, 1);
    }

    /* generate the increment code */
    putcbyte(c, OP_ADD);
    inst = putcbyte(c, OP_BR);
    putcword(c, test - inst - 3);

    /* branch to the loop body */
    fixupbranch(c, body, codeaddr(c));
    Require(c, tkn, T_EOL);
}

/* ParseNext - parse the 'NEXT' statement */
static void ParseNext(ParseContext *c)
{
    ParseTreeNode *var;
    int inst;
    switch (CheckBlock(c)) {
    case BLOCK_FOR:
        FRequire(c, T_IDENTIFIER);
        var = GetSymbolRef(c, c->token);
        /* BUG: check to make sure it matches the symbol used in the FOR */
        FreeParseTree(var);
        inst = putcbyte(c, OP_BR);
        putcword(c, c->bptr->ForBlock.nxt - inst - 3);
        fixupbranch(c, c->bptr->ForBlock.end, codeaddr(c));
        PopBlock(c);
        break;
    default:
        ParseError(c, "NEXT without a matching FOR");
        break;
    }
    FRequire(c, T_EOL);
}

/* ParseDo - parse the 'DO' statement */
static void ParseDo(ParseContext *c)
{
    PushBlock(c);
    c->bptr->type = BLOCK_DO;
    c->bptr->DoBlock.nxt = codeaddr(c);
    c->bptr->DoBlock.end = 0;
    FRequire(c, T_EOL);
}

/* ParseDoWhile - parse the 'DO WHILE' statement */
static void ParseDoWhile(ParseContext *c)
{
    PushBlock(c);
    c->bptr->type = BLOCK_DO;
    c->bptr->DoBlock.nxt = codeaddr(c);
    ParseRValue(c);
    putcbyte(c, OP_BRF);
    c->bptr->DoBlock.end = putcword(c, 0);
    FRequire(c, T_EOL);
}

/* ParseDoUntil - parse the 'DO UNTIL' statement */
static void ParseDoUntil(ParseContext *c)
{
    PushBlock(c);
    c->bptr->type = BLOCK_DO;
    c->bptr->DoBlock.nxt = codeaddr(c);
    ParseRValue(c);
    putcbyte(c, OP_BRT);
    c->bptr->DoBlock.end = putcword(c, 0);
    FRequire(c, T_EOL);
}

/* ParseLoop - parse the 'LOOP' statement */
static void ParseLoop(ParseContext *c)
{
    int inst;
    switch (CheckBlock(c)) {
    case BLOCK_DO:
        inst = putcbyte(c, OP_BR);
        putcword(c, c->bptr->DoBlock.nxt - inst - 3);
        fixupbranch(c, c->bptr->DoBlock.end, codeaddr(c));
        PopBlock(c);
        break;
    default:
        ParseError(c, "LOOP without a matching DO");
        break;
    }
    FRequire(c, T_EOL);
}

/* ParseLoopWhile - parse the 'LOOP WHILE' statement */
static void ParseLoopWhile(ParseContext *c)
{
    int inst;
    switch (CheckBlock(c)) {
    case BLOCK_DO:
        ParseRValue(c);
        inst = putcbyte(c, OP_BRT);
        putcword(c, c->bptr->DoBlock.nxt - inst - 3);
        fixupbranch(c, c->bptr->DoBlock.end, codeaddr(c));
        PopBlock(c);
        break;
    default:
        ParseError(c, "LOOP without a matching DO");
        break;
    }
    FRequire(c, T_EOL);
}

/* ParseLoopUntil - parse the 'LOOP UNTIL' statement */
static void ParseLoopUntil(ParseContext *c)
{
    int inst;
    switch (CheckBlock(c)) {
    case BLOCK_DO:
        ParseRValue(c);
        inst = putcbyte(c, OP_BRF);
        putcword(c, c->bptr->DoBlock.nxt - inst - 3);
        fixupbranch(c, c->bptr->DoBlock.end, codeaddr(c));
        PopBlock(c);
        break;
    default:
        ParseError(c, "LOOP without a matching DO");
        break;
    }
    FRequire(c, T_EOL);
}

/* ParseStop - parse the 'STOP' statement */
static void ParseStop(ParseContext *c)
{
    putcbyte(c, OP_HALT);
    FRequire(c, T_EOL);
}

/* Print - handle the 'PRINT' or 'DISPLAY' statements */
static void Print(ParseContext *c, uint8_t dev)
{
    int needNewline = VMTRUE;
    ParseTreeNode *expr;
    Token tkn;

    while ((tkn = GetToken(c)) != T_EOL) {
        switch (tkn) {
        case ',':
            needNewline = VMFALSE;
            putcbyte(c, OP_PRINTT);
            putcbyte(c, dev);
            break;
        case ';':
            needNewline = VMFALSE;
            break;
        default:
            needNewline = VMTRUE;
            SaveToken(c, tkn);
            expr = ParseExpr(c);
            code_rvalue(c, expr);
            switch (NodeType(c, expr)->id) {
            case TYPE_STRING:
                putcbyte(c, OP_PRINTS);
                putcbyte(c, dev);
                break;
            case TYPE_INTEGER:
            case TYPE_BYTE:
                putcbyte(c, OP_PRINTI);
                putcbyte(c, dev);
                break;
            default:
                ParseError(c, "invalid argument to %s", dev == 0 ? "PRINT" : "DISPLAY");
                break;
            }
            FreeParseTree(expr);
            break;
        }
    }

    if (needNewline) {
        putcbyte(c, OP_PRINTN);
        putcbyte(c, dev);
    }
}

/* ParsePrint - parse the 'PRINT' statement */
static void ParsePrint(ParseContext *c)
{
    Print(c, 0);
}

/* ParseWrite - parse the 'WRITE' statement */
static void ParseWrite(ParseContext *c)
{
}

/* ParseDisplay - parse the 'DISPLAY' statement */
static void ParseDisplay(ParseContext *c)
{
    Print(c, 1);
}

/* ParseInput - parse the 'INPUT' statement */
static void ParseInput(ParseContext *c)
{
    int dev = 0;
    Token tkn;
    PVAL pv;

    /* parse the prompt string */
    if ((tkn = GetToken(c)) == T_STRING) {
        putcbyte(c, OP_LIT);
        putcword(c, AddStringRef(AddString(c, c->token), codeaddr(c)));
        putcbyte(c, OP_PRINTS);
        putcbyte(c, dev);
        FRequire(c, ';');
    }

    /* force reading a new line */
    putcbyte(c, OP_GETLINE);
    putcbyte(c, dev);

    /* parse each input variable */
    if ((tkn = GetToken(c)) == T_IDENTIFIER) {
        SaveToken(c, tkn);
        do {
            ParseTreeNode *expr;
            FRequire(c, T_IDENTIFIER);
            expr = ParsePossibleArrayReference(c, GetSymbolRef(c, c->token));
            code_lvalue(c, expr, &pv);
            switch (NodeType(c, expr)->id) {
            case TYPE_STRING:
                putcbyte(c, OP_INPUTS);
                putcbyte(c, dev);
                break;
            case TYPE_INTEGER:
            case TYPE_BYTE:
                putcbyte(c, OP_INPUTI);
                putcbyte(c, dev);
                break;
            default:
                ParseError(c, "invalid argument to INPUT");
                break;
            }
            (*pv.fcn)(c, PV_STORE, &pv);
            FreeParseTree(expr);
        } while ((tkn = GetToken(c)) == ',');
    }
}

/* ParseOpen - parse the 'OPEN' statement */
static void ParseOpen(ParseContext *c)
{
}

/* ParseClose - parse the 'CLOSE' statement */
static void ParseClose(ParseContext *c)
{
}

/* ParseGoto - parse the 'GOTO' statement */
static void ParseGoto(ParseContext *c)
{
    FRequire(c, T_IDENTIFIER);
    putcbyte(c, OP_BR);
    putcword(c, ReferenceLabel(c, c->token, codeaddr(c)));
    FRequire(c, T_EOL);
}

/* DefineLabel - define a local label */
static void DefineLabel(ParseContext *c, char *name, int offset)
{
    Label *label;
    
    /* check to see if the label is already in the table */
    for (label = c->labels; label != NULL; label = label->next)
        if (strcasecmp(name, label->name) == 0) {
            if (!label->fixups)
                ParseError(c, "duplicate label: %s", name);
            else {
                fixupbranch(c, label->fixups, offset);
                label->offset = offset;
                label->fixups = 0;
            }
            return;
        }

    /* allocate the label structure */
    label = (Label *)Alloc("LBL", sizeof(Label) + strlen(name));
    memset(label, 0, sizeof(Label));
    strcpy(label->name, name);
    label->offset = offset;
    label->next = c->labels;
    c->labels = label;
}

/* ReferenceLabel - add a reference to a local label */
static int ReferenceLabel(ParseContext *c, char *name, int offset)
{
    Label *label;
    
    /* check to see if the label is already in the table */
    for (label = c->labels; label != NULL; label = label->next)
        if (strcasecmp(name, label->name) == 0) {
            int link;
            if (!(link = label->fixups))
                return label->offset - offset - 2;
            else {
                label->fixups = offset;
                return link;
            }
        }

    /* allocate the label structure */
    label = (Label *)Alloc("LBL", sizeof(Label) + strlen(name));
    memset(label, 0, sizeof(Label));
    strcpy(label->name, name);
    label->fixups = offset;
    label->next = c->labels;
    c->labels = label;

    /* return zero to terminate the fixup list */
    return 0;
}

/* CheckLabels - check for undefined labels and free the labels */
static void CheckLabels(ParseContext *c)
{
    Label *label, *next;
    for (label = c->labels; label != NULL; label = next) {
        next = label->next;
        if (label->fixups)
            Fatal("undefined label: %s", label->name);
        Free(label);
    }
    c->labels = NULL;
}

/* CheckBlock - make sure there is a block on the stack */
static BlockType CheckBlock(ParseContext *c)
{
    return c->bptr < c->blockBuf ? BLOCK_NONE : c->bptr->type;
}

/* PushBlock - push a block on the block stack */
static void PushBlock(ParseContext *c)
{
    if (++c->bptr >= c->btop)
        Fatal("statements too deeply nested");
}

/* PopBlock - pop a block off the block stack */
static void PopBlock(ParseContext *c)
{
    --c->bptr;
}
