/* db_vmflash.c - flash i/o routines
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *

Sample code sequence:

    // initialize the interpreter
    VM_InitImage();
    InitInterpreter(i, TARGET_BITMAP);
    
    // then call one of these to load an image
    LoadImageFile(i, "autorun.bai");
    InitMemoryImage(i);
    
    // execute the loaded code
    Execute(i);

 */

#include "../runtime/db_vm.h"
#include "../common/db_packet.h"

#define FLASH_PAGEOFFSET    (FLASH_PAGESIZE - 1)
#define FLASH_PAGEMASK      (~FLASH_PAGEOFFSET)
#define FLASH_ROUND(n)      (((int)(n) + FLASH_PAGESIZE - 1) & FLASH_PAGEMASK)

/* code space in flash */
static FileHdr *vmimage;
static uint8_t *vmcode;
static FLASH_SPACE uint8_t vmcodebuf[VMCODE_SIZE + 2 * FLASH_PAGESIZE - 1] = { 0 };

/* data space in sram */
static uint8_t vmdata[VMDATA_SIZE] VMDATA_ATTR;
static uint8_t *vmfree;
static unsigned int vmfreesize;

/* flash write buffer */
static uint8_t flashPageBuf[FLASH_PAGESIZE - PKTMAXLEN];
static unsigned int flashPage = 0;

/* local function prototypes */
static void WriteFlash(uint8_t *dst, uint8_t *src, int cnt);
static void FlushFlashBuffer(void);
static void WriteFlashPage(uint8_t *src, int cnt) FLASH_BOOT;

/* VM_InitImage - initialize the image buffer */
void VM_InitImage(void)
{
	// make sure the code starts on a flash page boundry
	vmimage = (FileHdr *)FLASH_ROUND(vmcodebuf);
    vmcode = (uint8_t *)vmimage + FLASH_PAGESIZE;

    // everything is free to start
    vmfree = vmdata;
    vmfreesize = VMDATA_SIZE;
}

/* VM_StoreMemoryImageHeader - store the header of an image in memory */
void VM_StoreMemoryImageHeader(Interpreter *i)
{
    FlushFlashBuffer();
    WriteFlash((uint8_t *)vmimage, (uint8_t *)&i->hdr, sizeof(FileHdr));
    FlushFlashBuffer();
}

/* VM_AllocateImage - allocate space for an image */
int VM_AllocateImage(Interpreter *i, uint16_t codeSize, uint16_t initSize, uint16_t dataSize, uint16_t heapSize)
{
    /* round up the code size to a flash page boundry */
    codeSize = FLASH_ROUND(codeSize);
    
    /* make sure there is enough memory */
    if (codeSize + initSize > VMCODE_SIZE || dataSize + heapSize > VMDATA_SIZE)
        return VMFALSE;

    /* setup the code and data space */
    i->code = vmcode;
    i->variables = (int16_t *)vmdata;

    /* setup the heap space */
    if (heapSize)
        i->heap = (StringHeap *)(vmdata + dataSize);

    /* setup free data space */
    vmfree = vmdata + dataSize + heapSize;
    vmfreesize = VMDATA_SIZE - dataSize - heapSize;

    /* initialize the flash page write buffer */
    flashPage = 0;

    /* return successfully */
    return VMTRUE;
}

/* VM_FreeSpace - get the remaining free data space */
uint8_t *VM_FreeSpace(int *pSize)
{
    return (*pSize = vmfreesize) > 0 ? vmfree : NULL;
}

/* VM_FreeImage - free an image */
void VM_FreeImage(Interpreter *i)
{
}

/* VM_LoadCode - load code into flash */
void VM_LoadCode(uint16_t dstoff, uint8_t *src, int cnt)
{
    WriteFlash(vmcode + dstoff, src, cnt);
}

/* VM_LoadInit - load initialization data into flash */
void VM_LoadInit(uint16_t dstoff, uint8_t *src, int cnt)
{
    uint8_t *init = vmcode + FLASH_ROUND(interp.hdr.codeSize);
    WriteFlash(init + dstoff, src, cnt);
}

/* VM_ReadInit - read initialization data from flash */
void VM_ReadInit(uint16_t off, uint8_t *buf, int cnt)
{
    uint8_t *p = vmcode + FLASH_ROUND(interp.hdr.codeSize) + off;
    while (--cnt >= 0)
        *buf++ = VMCODEBYTE(p++);
}

/* WriteFlash - write a block of memory to flash
  
  Parameters:
    dst:            address in flash memory on a PKTMAXLEN byte boundry
    src:            the address of the data to write
    cnt:            number of bytes in the src buffer (must be <= PKTMAXLEN)

*/
static void WriteFlash(uint8_t *dst, uint8_t *src, int cnt)
{
    unsigned int page = (unsigned int)dst & FLASH_PAGEMASK;
    uint16_t pageoff = (unsigned int)dst & FLASH_PAGEOFFSET;
    
    /* flush the buffer if the page in the buffer is different from this page */
    if (flashPage && flashPage != page)
        FlushFlashBuffer();

    /* buffer the data this isn't the last PKTMAXLEN bytes of the page */
    if (pageoff < sizeof(flashPageBuf)) {
        memcpy(&flashPageBuf[pageoff], src, cnt);
        if (cnt < PKTMAXLEN)
            memset(&flashPageBuf[pageoff] + cnt, 0, PKTMAXLEN - cnt);
        flashPage = page;
    }

    /* write the flash page if this is in the last PKTMAXLEN bytes */
    else {
        WriteFlashPage(src, cnt);
        flashPage = 0;
    }
}

/* FlushFlashBuffer - flush any data currently in the flash page buffer */
static void FlushFlashBuffer(void)
{
    if (flashPage) {
        WriteFlashPage(NULL, 0);
        flashPage = 0;
    }
}

/* WriteFlashPage - write a single page to flash

  Parameters:
    src:            the address of the data that will occupy the remaining PKTMAXLEN bytes of the page
    cnt:            number of bytes in the src buffer (must be <= PKTMAXLEN)

  Globals:
    flashPage:      the address of the first byte of a flash page
    flashPageBuf:   buffer containing the first FLASH_PAGESIZE - PKTMAXLEN bytes of page data

*/

/* VM_LoadMemoryImageHeader - load the header of an image already in memory

*/

#ifdef WIN32

static void WriteFlashPage(uint8_t *src, int cnt)
{
    memcpy((void *)flashPage, flashPageBuf, sizeof(flashPageBuf));
    if (src)
        memcpy((void *)(flashPage + sizeof(flashPageBuf)), src, cnt);
}

#endif  // WIN32

#ifdef XGSAVR

int VM_LoadMemoryImageHeader(Interpreter *i)
{
    memcpy_P(&i->hdr, vmimage, sizeof(FileHdr));
    return VMTRUE;
}

static void WriteFlashPage(uint8_t *src, int cnt)
{
    uint16_t wdst, end;
    uint16_t *wsrc;
    
    /* disable interrupts */
    cli();

    /* wait for any eeprom operations to complete */
    eeprom_busy_wait();

    /* copy the data already in the buffer */
    end = flashPage + sizeof(flashPageBuf);
    for (wdst = flashPage, wsrc = (uint16_t *)flashPageBuf; wdst < end; wdst += 2)
        boot_page_fill(wdst, *wsrc++);

    /* copy the data we are writing */
    for (wsrc = (uint16_t *)src; (cnt -= 2) >= 0; wdst += 2)
        boot_page_fill(wdst, *wsrc++);

    /* zero the rest of the page */
    end = flashPage + FLASH_PAGESIZE;
    for (; wdst < end; wdst += 2)
        boot_page_fill(wdst, 0);

    /* erase the page */
    boot_page_erase(flashPage);
    boot_spm_busy_wait();

    /* write the updated page */
    boot_page_write(flashPage);
    boot_spm_busy_wait();

    /* enable the read-while-write memory */
    boot_rww_enable();

    /* enable the interrupts */
    sei();
}

#endif  // XGSAVR

#ifdef XGSPIC

#define Get24BitAddress(a)  ((PSVPAG << 15) | ((a) & 0x7fff))

int VM_LoadMemoryImageHeader(Interpreter *i)
{
    memcpy(&i->hdr, vmimage, sizeof(FileHdr));
    return VMTRUE;
}

static void WriteFlashPage(uint8_t *src, int cnt)
{
    _prog_addressT page = Get24BitAddress(flashPage);
    _prog_addressT wdst, end;
    int *wsrc;

    /* erase the page */
    _erase_flash(page);

    /* copy the data already in the buffer */
    end = page + sizeof(flashPageBuf);
    wsrc = (int *)flashPageBuf;
    for (wdst = page; wdst < end; wdst += FLASH_ROWSIZE) {
        _write_flash16(wdst, wsrc);
        wsrc += _FLASH_ROW;
    }

    /* copy the data we are writing */
    wsrc = (int *)src;
    for (; cnt >= FLASH_ROWSIZE; cnt -= FLASH_ROWSIZE) {
        _write_flash16(wdst, wsrc);
        wsrc += _FLASH_ROW;
        wdst += FLASH_ROWSIZE;
    }

    /* copy the remaining partial row */
    if (cnt > 0) {
        memcpy(flashPageBuf, src, cnt);
        memset(&flashPageBuf[cnt], 0, FLASH_ROWSIZE - cnt);
        _write_flash16(wdst, (int *)flashPageBuf);
    }
}

#endif  // XGSPIC
