/* db_vmheap.c - string heap manager
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <string.h>
#include "db_vm.h"

/* round to multiple of the word size */
#define WORDSIZE(n)     (((n) + 1) & ~1)

/* InitHeap - initialize a string heap */
void InitHeap(Interpreter *i, uint16_t size, uint16_t nHandles)
{
    uint16_t dataSize = size - sizeof(StringHeap) - nHandles * sizeof(StringHandle);
    StringHandle *handle;
    uint8_t *data;
    uint16_t cnt;

    /* setup the heap header */
    data = (uint8_t *)i->heap;
    data += WORDSIZE(sizeof(StringHeap));
    size -= WORDSIZE(sizeof(StringHeap));

    /* setup the handle array */
    i->heap->handles = (StringHandle *)(data + dataSize);
    i->heap->endHandles = i->heap->handles + nHandles;
    i->heap->nHandles = nHandles;
    
    /* create the handle free list */
    i->heap->freeHandles = NULL;
    for (handle = i->heap->handles, cnt = i->heap->nHandles; cnt > 0; ++handle, --cnt) {
        handle->data = (uint8_t *)i->heap->freeHandles;
        i->heap->freeHandles = handle;
    }

    /* setup the string heap */
    i->heap->data = i->heap->free = data;
}

/* StringAlloc - allocate a new string */
int16_t StringAlloc(Interpreter *i, uint16_t length)
{
    uint16_t size = sizeof(StringHdr) + WORDSIZE(length);
    StringHandle *h;
    StringHdr *hdr;

    /* make sure a heap was configured */
    if (!i->heap)
        Abort(i, "no string heap available");

    /* find a free handle */
    if (!(h = i->heap->freeHandles))
        Abort(i, "insufficient string heap handles");

    /* remove the handle from the free list */
    i->heap->freeHandles = (StringHandle *)h->data;

    /* make sure there's enough space */
    if ((uint8_t *)i->heap->handles - i->heap->free < size) {
        CompactHeap(i);
        if ((uint8_t *)i->heap->handles - i->heap->free < size) {
            h->data = (uint8_t *)i->heap->freeHandles;
            i->heap->freeHandles = h;
            Abort(i, "insufficient string heap space");
        }
    }

    /* allocate the next available block */
    hdr = (StringHdr *)i->heap->free;
    i->heap->free += size;
    
    /* store the backpointer and length */
    hdr->handle = h;
    hdr->length = length;
    hdr->refCnt = 1;
    
    /* store the pointer from the handle to the heap data */
    h->data = (uint8_t *)(hdr + 1);

    /* return successfully */
    return i->heap->endHandles - h;
}

/* StringAddRef - add a reference to a string */
void StringAddRef(Interpreter *i, int16_t handle)
{
    if (handle && !(handle & 0x8000)) {
        StringHandle *h = i->heap->endHandles - handle;
        StringHdr *hdr = (StringHdr *)(h->data - sizeof(StringHdr));
        ++hdr->refCnt;
    }
}

/* StringRelease - release a reference to a string */
void StringRelease(Interpreter *i, int16_t handle)
{
    if (handle && !(handle & 0x8000)) {
        StringHandle *h = i->heap->endHandles - handle;
        StringHdr *hdr = (StringHdr *)(h->data - sizeof(StringHdr));
        if (--hdr->refCnt == 0) {
            h->data = (uint8_t *)i->heap->freeHandles;
            i->heap->freeHandles = h;
            hdr->handle = NULL;
        }
    }
}

/* GetStringPtr - get a string argument that could be in code space or in the heap */
char *GetStringPtr(Interpreter *i, int16_t value, int *pInHeap)
{
    if (value & 0x8000) {
        *pInHeap = VMFALSE;
        return (char *)i->code + (value & 0x7fff);
    }
    *pInHeap = VMTRUE;
    return value ? (char *)GetHeapStringPtr(i, value) : "";
}

/* GetStringLen - get a string argument that could be in code space or in the heap */
int GetStringLen(Interpreter *i, int16_t value)
{
	if (value & 0x8000)
        return VMSTRLEN((char *)i->code + (value & 0x7fff));
    return value ? (int)GetHeapStringLen(i, value) : 0;
}

/* CompactHeap - compact the heap */
void CompactHeap(Interpreter *i)
{
    if (i->heap) {
        uint8_t *data, *next, *free;

        /* initialize the heap scan */
        data = next = i->heap->data;
        free = i->heap->free;
    
        /* compact the heap */
        while (data < free) {
            StringHdr *hdr = (StringHdr *)data;
            uint16_t size = sizeof(StringHdr) + WORDSIZE(hdr->length);
            if (hdr->handle) {
                if (data != next) {
                    hdr->handle->data = next + sizeof(StringHdr);
                    memmove(next, data, size);
                }
                next += size;
            }
            data += size;
        }
    
        /* store the new free pointer */
        i->heap->free = next;
    }
}

/* DumpHeap - dump the heap */
void DumpHeap(Interpreter *i)
{
    if (i->heap) {
        uint8_t *data, *free;

        /* initialize the heap scan */
        data = i->heap->data;
        free = i->heap->free;
    
        /* compact the heap */
        while (data < free) {
            StringHdr *hdr = (StringHdr *)data;
            uint16_t size = sizeof(StringHdr) + WORDSIZE(hdr->length);
            uint8_t *addr = (uint8_t *)(hdr + 1);
            int length = hdr->length;
            VM_printf("%04x h: %04x, c: %d, l: %d, d: %04x '",
                      data - i->heap->data,
                      hdr->handle ? i->heap->endHandles - hdr->handle : 0,
                      hdr->refCnt,
                      length,
                      addr - i->heap->data);
            while (--length >= 0)
                VM_putchar(*addr++);
            VM_printf("'\n");
            data += size;
        }
    }
}
