#include "XGS_AVR_SYSTEM_V010.h"            

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include "../common/db_packet.h"

//#define ENABLE_LOGGING

#define RX_BUFFER_SIZE	135
#define TX_BUFFER_SIZE	32

typedef struct {
    volatile uint8_t head;
    volatile uint8_t tail;
    uint8_t *data;
    uint8_t size;
	int overwrite;
} Buffer;

static Buffer rx_buf;
static uint8_t rx_data[RX_BUFFER_SIZE + 1];
static Buffer tx_buf;
static uint8_t tx_data[TX_BUFFER_SIZE + 1];

static volatile uint16_t timerCount;

static void BufInit(Buffer *buf, uint8_t *data, uint8_t size, int overwrite);
static int BufPutByte(Buffer *buf, int byte);
static int BufGetByte(Buffer *buf);

static int UART_init(uint16_t baudrate);
static void UART_term(void);
int UART_putc(int ch);
int UART_getc(void);

static uint16_t ReadTimer(void);

int md_init(char *port)
{
	BufInit(&rx_buf, rx_data, sizeof(rx_data), VMTRUE);
	BufInit(&tx_buf, tx_data, sizeof(tx_data), VMFALSE);

	UART_init(9600);
	
	TCCR0A = 0;
	TCCR0B = (1 << CS02) | (1 << CS00);	// prescale 1024
	TIFR0 = 1 << TOV0;
	TIMSK0 = 1 << TOIE0;
	timerCount = 0;

#ifdef ENABLE_LOGGING
    FSremove("mdio.log");
#endif

	sei();

    return VMTRUE;
}

void md_done(void)
{
	TIMSK0 = 0 << TOIE2;
	UART_term();
}

int md_get(int timeout)
{
    uint16_t startTime = ReadTimer();
    timeout *= F_CPU / (1024UL * 256UL);
    for (;;) {
        int ch;
        if ((ch = UART_getc()) != -1)
            return ch;
        if (ReadTimer() - startTime >= timeout)
            return MD_TIMEOUT;
    }
}

int md_write(uint8_t *buf, int len)
{
    while (--len >= 0)
        UART_putc(*buf++);
    return VMTRUE;
}

int UART_putc(int ch)
{
	/* wait for space */
	while (!BufPutByte(&tx_buf, ch))
		;
	
	/* enable the buffer empty interrupt */
	UCSR0B |= 1 << UDRIE0;

	/* return the character */
	return ch;
}

int UART_getc(void)
{
    int byte;
	
    /* disable the receive interrupt */
    UCSR0B &= ~(1 << RXCIE0);

	/* get a character from the buffer */
	byte = BufGetByte(&rx_buf);

    /* enable the receive interrupt */
    UCSR0B |= 1 << RXCIE0;

	/* return the character */
	return byte;
}

static void BufInit(Buffer *buf, uint8_t *data, uint8_t size, int overwrite)
{
    buf->head = buf->tail = 0;
    buf->data = data;
    buf->size = size;
	buf->overwrite = overwrite;
}

static int BufPutByte(Buffer *buf, int byte)
{
    uint8_t head, tail;
    
    /* advance the write pointer */
    if ((head = buf->head + 1) >= buf->size)
        head = 0;
    
    /* check for the buffer being full */
    if (head == buf->tail) {

		/* overwrite the oldest data */
		if (buf->overwrite) {
			if ((tail = buf->tail + 1) >= buf->size)
				tail = 0;
			buf->tail = tail;
		}

		/* return failure */
		else
        	return VMFALSE;
	}
        
    /* add the new byte to the buffer */
    buf->data[buf->head] = byte;
    buf->head = head;
    
    /* return successfully */
    return VMTRUE;
}

static int BufGetByte(Buffer *buf)
{
    uint8_t tail;
    int byte;

    /* check for the buffer being empty */
    if (buf->head == buf->tail)
        return -1;
        
    /* advance the write pointer */
    if ((tail  = buf->tail + 1) >= buf->size)
        tail = 0;
    
    /* add the new byte to the buffer */
    byte = buf->data[buf->tail];
    buf->tail = tail;
    
    /* return successfully */
    return byte;
}

/* the following function was taken directly from Andre's XGS_AVR_UART_DRV_V010.c file */

static int UART_init(uint16_t baudrate)
{
// initialize the UART to N81 data format, enables interrupts etc. and computes baudrate register value
// note 1: we do not need to initialize the port pins used by the UART since enabling overrides any settings we might
// assign, but it can't hurt to enable the TX to output and RX to input at some point...

// note 2: due to the overclocking of the AVR at @28.63636, the usual formula to compute the baud rate register isn't 
// accurate enough, therefore, constants are used instead. If you slow the processor back down to 20 Mhz or under then 
// the formula will work again
// baudrate_reg = ((F_CPU)/((baudrate)*16))-1;

uint16_t baudrate_reg;

// run this code only for 8*NTSC overclock modes, else try and use general formula
if (F_CPU == 28636360UL)
    {
    // based on sent baud rate compute baudrate register value
    switch( baudrate )
        {
        case 9600: // this range works [176..188]
            {
            baudrate_reg = 182; 
            } break;

        case 19200: // this range works [88..90]
            {
            baudrate_reg = 89; 
            } break;

        case 38400: // this range works [45], 38400 is the absolute limit you can run the overclocked UART at
            {       // if you can't tolerate any errors, drop back to 19200, else 38400 is 99.9% error free
            baudrate_reg = 45; 
            } break;
        
        default:  // caller is requesting a non-computed baud rate, use formula and hope for the best
            {
            baudrate_reg = ((F_CPU)/((baudrate)*16))-1;
            } break;

        } // end switch
    } // end if
else // use formula
    {
    // best guess
    baudrate_reg = ((F_CPU)/((baudrate)*16))-1;
    } // end else

// UBRR0 = baudrate;

// UART baud rate register
UBRR0H = (unsigned char)(baudrate_reg >> 8);
UBRR0L = (unsigned char)(baudrate_reg);

// table of various experimental settings, format: "[ baud rate register ] @clock / baudrate"
// [22] @14.31818 / 38400, [14]  @14.31818 / 57600,  [7] @14.31818 / 115200, 
// [11] @21.4775 / 115200,  [33] @21.4775 / 38400, 
// [45] @28.63636 / 38400, [88..90] @28.63636 / 19200, [176..188] @28.63636 / 9600
// [216] @ 33.3333 / 9600
                        
// UART control register
// enable receiver, transmitter and receive interrupt
UCSR0B = (1 << RXEN0) | (1 << TXEN0) | (1 << RXCIE0); 

// now set data frame format to N:8:1, async mode
// 168, 644
UCSR0C = (0 << USBS0) | ( 1 << UCSZ01) | (1 << UCSZ00) | (0 << UMSEL00) | (0 << UMSEL01) | (0 << UPM01) | ( 0 << UPM00 );

// 162
//UCSR0C = (0<<USBS0)|(1<<UCSZ01)|(1<<UCSZ00)|(0<<UMSEL0)|(0<<UMSEL1)|(0<<UPM01)|(0<<UPM00);

// return success, return better error code later
return(1);

} // end UART_Init

static void UART_term(void)
{
	UCSR0B = (0 << RXEN0) | (0 << TXEN0) | (0 << RXCIE0) | (0 << UDRIE0); 
}

ISR(USART0_RX_vect)
{
	BufPutByte(&rx_buf, UDR0);
}

ISR(USART0_UDRE_vect)
{
	int byte;
	if ((UCSR0A & (1 << UDRE0)) && (byte = BufGetByte(&tx_buf)) != -1)
		UDR0 = byte;
	else
	    UCSR0B &= ~(1 << UDRIE0);
}

static uint16_t ReadTimer(void)
{
	uint16_t now;
	TIMSK0 = 0 << TOIE2;
	now = timerCount;
	TIMSK0 = 1 << TOIE2;
	return now;
}

ISR(TIMER0_OVF_vect)
{
	++timerCount;
}

