///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// Copyright Nurve Networks LLC 2008
// 
// Filename: REM_AVR_PARROT_NTSC_01.c
//
// Original Author: Remi Veilleux
// 
// Last Modified: 12.22.08
//
// Overview: "true-color" image output with a 224x168 resolution
//
// Controls: gamepad, use d-pad (up/down, left/right) to distort the image with a smooth sinusoid
// The effect is running at a constant 60 frame per second refresh rate (NTSC)
// 
// 
// Demo must be linked with following source files:
// 
// XGS_AVR_SYSTEM_V010.c              - System variables.
// REM_AVR_NTSC_BMP_224_168_8_V010.s  - ASM NTSC bitmap driver.
// XGS_AVR_GAMEPAD_DRV_V010.c         - Gamepad driver.
//
// Note 1: header files for all sources must be in current working directory, so compiler can find them
//
// 
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES ///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// define chip type here
#define __AVR_ATmega644__

// define CPU frequency in Mhz here if not defined in Makefile, this is used for conditional compilation, so needs
// to be first, before other XGS includes, also note there is a copy of this constant in XGS_AVR_SYSTEM_V010, so if you 
// change it here, also change it in the header to make certain everthing is sync'ed
#define F_CPU 28636360UL

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <avr/eeprom.h>
#include <avr/io.h>
#include <avr/sleep.h>

#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <compat/twi.h>

// include local XGS API header files
#include "XGS_AVR_SYSTEM_V010.h"
#include "XGS_AVR_GAMEPAD_DRV_V010.h"

// include header file for video driver
#include "REM_AVR_NTSC_BMP_224_168_8_V010.h"

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// TYPES  /////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
extern UCHAR overscan_color;
extern UCHAR hscroll;
extern UCHAR vscroll;
extern UINT framecount;

// NOTE: these externs need to be declared volatile because they are modified by the asm interrupt driver
extern volatile unsigned int  curr_line;
extern volatile unsigned char curr_raster_line;


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// GLOBALS/////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MAIN ///////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


void GFX_Init(int video_interrupt_rate)
{
	// this function initializes the I/O ports for NTSC/PAL/VGA graphics as well as initilizes variables and turns on the 
	// driver interrupt which will immediately start feeding the data sources to the currently installed ASM driver, thus
	// make sure any pointers to data structures are initialized before calling this function
	// when done with graphics and you want to stop calling the graphics interrupt, call GFX_Shutdown(..)
	//
	// Inputs:
	//     video_interrupt_rate - the clock divider, so that the number of clocks per interrupt is equal to the HSYNC rate
	//                            since the interrupt handler runs every HSYNC in both NTSC/PAL and VGA modes

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// VIDEO INTERRUPT EXPERIMENT /////////////////////////////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// main graphics port that luma/chroma and VGA signals come from, port A
	#define GFX_MAIN_PORT       PORTA
	#define GFX_MAIN_DDR        DDRA

	// the "saturation" control is connected to port B
	#define GFX_SATURATION_PORT PORTB
	#define GFX_SATURATION_DDR  DDRB

	// set video ports to output
	GFX_MAIN_DDR  = 0xFF; // NTSC/PAL LUMA7:4 | CHROMA3:0, VGA7:0 [vsync, hsync, B1,B0, G1,G0, G1,G0] 
	              //                                        b7      b6    b5:4   b3:2   b1:0
	GFX_MAIN_PORT = 0x00;

	// set sat control bits PB2, PB3 as outputs
	GFX_SATURATION_DDR  =  SETPORTBITS(0,0,0,0,1,1,0,0);
	PORTB               =  SETPORTBITS(0,0,0,0,1,1,0,0);

	// set up timer                                     
	TCCR1A = (0 << COM1A1) | (0 << COM1A0) | (0 << COM1B1) | (0 << COM1B0) | (0 << WGM11) | (0 << WGM10);

	// set clock pre-scaler and counter mode
	TCCR1B = (0 << WGM13) | (1 << WGM12) | (0 << CS12) | (0 << CS11)  | (1 << CS10);    

	// 168. 644 use TIMSK1,TIFR1
	// 162 use TIMSK,TIFR

	// enable timer interrupts for timer 
	TIMSK1 = (1 << OCIE1A);                  //(1<<TOIE1);                 

	// clear and pendings timer x CTC interrupts (1 clears it)
	TIFR1  = (1 << OCF1A); // (1<<TOV1);                  

	// set output compare register value
	OCR1A  = video_interrupt_rate; // 1817+-1 for NTSC at 28.636 Mhz = 15.748Hkz hsync rate, 910 for VGA @ 28.636Mhz = 31.468Khz, hsync rate

	// enable interrupt
	sei();
} // end GFX_Init

int main(void)
{
	int gamepad;           // used to read gamepad

	// initialize the video driver and GFX engine
	GFX_Init( VIDEO_INTERRUPT_RATE );

	// initialize gamepad
	Gamepad_Init();

	// set processor sleep mode
	set_sleep_mode( SLEEP_MODE_IDLE );
	sleep_enable();

	vscroll = 127;
	hscroll = 80;
	//overscan_color = NTSC_ORANGE;

	// main event loop
	while(1)
	{
	    // main loop that executes and is interrupted by video interrupt

	    // now goto sleep and wait for interrupt
	    sleep_cpu();

	    // only do video processing on a single line (at the top of the video scan)
	    if ( curr_line == END_ACTIVE_SCAN+1 )
	    {
	        // erase screen?
	        gamepad = Gamepad_Read(GAMEPAD0);

			if ((gamepad & GAMEPAD_UP) && vscroll > 0) vscroll--;
			if ((gamepad & GAMEPAD_DOWN) && vscroll < 255) vscroll++;
			if ((gamepad & GAMEPAD_LEFT) && hscroll > 0) hscroll--;
			if ((gamepad & GAMEPAD_RIGHT) && hscroll < 255) hscroll++;

			framecount++;
      
	    } // end if

	} // end while
} // end main

